from __future__ import annotations

from collections.abc import Mapping, Sequence

import rflx.expression as expr
from rflx.ada import (
    TRUE,
    Add,
    And,
    AndThen,
    Assignment,
    Call,
    CallStatement,
    Case,
    ChoiceList,
    Div,
    Equal,
    Expr,
    ExpressionFunctionDeclaration,
    First,
    ForIn,
    FormalSubprogramDeclaration,
    ForSomeIn,
    FunctionSpecification,
    Ghost,
    If,
    IfStatement,
    In,
    Indexed,
    InOutParameter,
    Last,
    Length,
    Less,
    LessEqual,
    Mod,
    Mul,
    NamedAggregate,
    Not,
    Number,
    ObjectDeclaration,
    Or,
    OutParameter,
    Parameter,
    Postcondition,
    Pragma,
    PragmaStatement,
    Precondition,
    ProcedureSpecification,
    Result,
    ReturnStatement,
    Selected,
    Size,
    Slice,
    String,
    Sub,
    SubprogramBody,
    SubprogramDeclaration,
    UnitPart,
    Variable,
)
from rflx.const import BUILTINS_PACKAGE
from rflx.identifier import ID
from rflx.model import FINAL, ByteOrder, Composite, Field, Message, Scalar, Type

from . import common, const


class ParserGenerator:
    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    def create_get_function(
        self,
        message: Message,
        scalar_fields: Mapping[Field, Scalar],
        composite_fields: Sequence[Field],
    ) -> UnitPart:
        if not scalar_fields:
            return UnitPart()

        comparison_to_aggregate = any(
            (isinstance(t, Composite) and common.has_aggregate_dependent_condition(message, f))
            for f, t in message.field_types.items()
        )

        big_endian_fields = [
            f for f in scalar_fields if message.byte_order[f] == ByteOrder.HIGH_ORDER_FIRST
        ]
        little_endian_fields = [
            f for f in scalar_fields if message.byte_order[f] == ByteOrder.LOW_ORDER_FIRST
        ]

        return UnitPart(
            [],
            [
                SubprogramBody(
                    FunctionSpecification(
                        "Get",
                        const.TYPES_BASE_INT,
                        [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
                    ),
                    (
                        [
                            *common.field_bit_location_declarations(Variable("Fld")),
                            ObjectDeclaration(
                                ["Buffer_First"],
                                const.TYPES_INDEX,
                                Call(const.TYPES_TO_INDEX, [Variable("First")]),
                                constant=True,
                            ),
                            ObjectDeclaration(
                                ["Buffer_Last"],
                                const.TYPES_INDEX,
                                Call(const.TYPES_TO_INDEX, [Variable("Last")]),
                                constant=True,
                            ),
                            ObjectDeclaration(
                                ["Offset"],
                                const.TYPES_OFFSET,
                                Call(
                                    const.TYPES_OFFSET,
                                    [
                                        Mod(
                                            Sub(
                                                Size(const.TYPES_BYTE),
                                                Mod(Variable("Last"), Size(const.TYPES_BYTE)),
                                            ),
                                            Size(const.TYPES_BYTE),
                                        ),
                                    ],
                                ),
                                constant=True,
                            ),
                            ObjectDeclaration(
                                ["Size"],
                                "Positive",
                                Case(
                                    Variable("Fld"),
                                    [
                                        *[
                                            (Variable(f.affixed_name), t.size.ada_expr())
                                            for f, t in scalar_fields.items()
                                        ],
                                        *(
                                            [(Variable("others"), Last("Positive"))]
                                            if composite_fields
                                            else []
                                        ),
                                    ],
                                ),
                                constant=True,
                            ),
                            ObjectDeclaration(
                                ["Byte_Order"],
                                const.TYPES_BYTE_ORDER,
                                (
                                    If(
                                        [
                                            (
                                                In(
                                                    Variable("Fld"),
                                                    ChoiceList(
                                                        *[
                                                            Variable(f.affixed_name)
                                                            for f in big_endian_fields
                                                        ],
                                                    ),
                                                ),
                                                Variable(const.TYPES_HIGH_ORDER_FIRST),
                                            ),
                                        ],
                                        Variable(const.TYPES_LOW_ORDER_FIRST),
                                    )
                                    if big_endian_fields and little_endian_fields
                                    else (
                                        Variable(const.TYPES_HIGH_ORDER_FIRST)
                                        if big_endian_fields
                                        else Variable(const.TYPES_LOW_ORDER_FIRST)
                                    )
                                ),
                                constant=True,
                            ),
                        ]
                        if scalar_fields or comparison_to_aggregate
                        else []
                    ),
                    [
                        ReturnStatement(
                            Call(
                                const.TYPES_OPERATIONS * "Extract",
                                [
                                    Variable("Ctx.Buffer.all"),
                                    Variable("Buffer_First"),
                                    Variable("Buffer_Last"),
                                    Variable("Offset"),
                                    Variable("Size"),
                                    Variable("Byte_Order"),
                                ],
                            ),
                        ),
                    ],
                    [
                        Precondition(
                            AndThen(
                                Call(
                                    self.prefix * message.identifier * "Has_Buffer",
                                    [Variable("Ctx")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Sufficient_Buffer_Length",
                                    [Variable("Ctx"), Variable("Fld")],
                                ),
                                *(
                                    [
                                        Not(
                                            Call(
                                                self.prefix
                                                * message.identifier
                                                * "Composite_Field",
                                                [Variable("Fld")],
                                            ),
                                        ),
                                    ]
                                    if composite_fields
                                    else []
                                ),
                            ),
                        ),
                    ],
                ),
            ],
        )

    def create_verify_procedure(
        self,
        message: Message,
        scalar_fields: Mapping[Field, Scalar],
        composite_fields: Sequence[Field],
    ) -> UnitPart:
        specification = ProcedureSpecification(
            "Verify",
            [InOutParameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        valid_field_condition = AndThen(
            Call(
                "Valid_Value",
                [Variable("Fld"), Variable("Value")],
            ),
            Call(
                "Field_Condition",
                [
                    Variable("Ctx"),
                    Variable("Fld"),
                    *([Variable("Value")] if common.has_value_dependent_condition(message) else []),
                    *(
                        [
                            Slice(
                                Variable("Ctx.Buffer.all"),
                                Call(
                                    const.TYPES_TO_INDEX,
                                    [Call("Field_First", [Variable("Ctx"), Variable("Fld")])],
                                ),
                                Call(
                                    const.TYPES_TO_INDEX,
                                    [Call("Field_Last", [Variable("Ctx"), Variable("Fld")])],
                                ),
                            ),
                        ]
                        if common.has_aggregate_dependent_condition(message)
                        else []
                    ),
                    *(
                        [Call("Field_Size", [Variable("Ctx"), Variable("Fld")])]
                        if common.has_size_dependent_condition(message)
                        else []
                    ),
                ],
            ),
        )

        last = Mul(
            Div(
                Add(
                    Call("Field_Last", [Variable("Ctx"), Variable("Fld")]),
                    Size(const.TYPES_BYTE),
                    -Number(1),
                ),
                Size(const.TYPES_BYTE),
            ),
            Size(const.TYPES_BYTE),
        )
        set_cursors_statements = [
            *(
                [
                    PragmaStatement(
                        "Assert",
                        [
                            If(
                                [
                                    (
                                        Or(
                                            *[
                                                Equal(Variable("Fld"), Variable(f.affixed_name))
                                                for f in message.direct_predecessors(FINAL)
                                            ],
                                        ),
                                        Equal(
                                            Mod(
                                                Call(
                                                    "Field_Last",
                                                    [Variable("Ctx"), Variable("Fld")],
                                                ),
                                                Size(const.TYPES_BYTE),
                                            ),
                                            Number(0),
                                        ),
                                    ),
                                ],
                            ),
                        ],
                    ),
                ]
                if len(message.fields) > 1
                else []
            ),
            # Improve provability of context predicate
            PragmaStatement("Assert", [Equal(Mod(last, Size(const.TYPES_BYTE)), Number(0))]),
            Assignment(Variable("Ctx.Verified_Last"), last),
            PragmaStatement(
                "Assert",
                [
                    LessEqual(
                        Call("Field_Last", [Variable("Ctx"), Variable("Fld")]),
                        Variable("Ctx.Verified_Last"),
                    ),
                ],
            ),
            (
                IfStatement(
                    [
                        (
                            Call("Composite_Field", [Variable("Fld")]),
                            [set_context_cursor_composite_field("Fld")],
                        ),
                    ],
                    [set_context_cursor_scalar()],
                )
                if scalar_fields and composite_fields
                else (
                    set_context_cursor_scalar()
                    if scalar_fields and not composite_fields
                    else set_context_cursor_composite_field("Fld")
                )
            ),
        ]

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            Call(
                                self.prefix * message.identifier * "Has_Buffer",
                                [Variable("Ctx")],
                            ),
                        ),
                        Postcondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                *common.context_invariant(message),
                            ),
                        ),
                    ],
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    [ObjectDeclaration(["Value"], const.TYPES_BASE_INT)],
                    [
                        IfStatement(
                            [
                                (
                                    AndThen(
                                        Call(
                                            "Invalid",
                                            [Indexed(Variable("Ctx.Cursors"), Variable("Fld"))],
                                        ),
                                        Call(
                                            "Valid_Next",
                                            [Variable("Ctx"), Variable("Fld")],
                                        ),
                                    ),
                                    [
                                        IfStatement(
                                            [
                                                (
                                                    Call(
                                                        "Sufficient_Buffer_Length",
                                                        [Variable("Ctx"), Variable("Fld")],
                                                    ),
                                                    [
                                                        Assignment(
                                                            "Value",
                                                            (
                                                                If(
                                                                    [
                                                                        (
                                                                            Call(
                                                                                "Composite_Field",
                                                                                [
                                                                                    Variable("Fld"),
                                                                                ],
                                                                            ),
                                                                            Number(0),
                                                                        ),
                                                                    ],
                                                                    Call(
                                                                        "Get",
                                                                        [
                                                                            Variable("Ctx"),
                                                                            Variable("Fld"),
                                                                        ],
                                                                    ),
                                                                )
                                                                if scalar_fields
                                                                and composite_fields
                                                                else (
                                                                    Call(
                                                                        "Get",
                                                                        [
                                                                            Variable("Ctx"),
                                                                            Variable("Fld"),
                                                                        ],
                                                                    )
                                                                    if scalar_fields
                                                                    else Number(0)
                                                                )
                                                            ),
                                                        ),
                                                        IfStatement(
                                                            [
                                                                (
                                                                    valid_field_condition,
                                                                    set_cursors_statements,
                                                                ),
                                                            ],
                                                            [
                                                                Assignment(
                                                                    Indexed(
                                                                        Variable("Ctx.Cursors"),
                                                                        Variable("Fld"),
                                                                    ),
                                                                    NamedAggregate(
                                                                        ("others", Variable("<>")),
                                                                    ),
                                                                ),
                                                            ],
                                                        ),
                                                    ],
                                                ),
                                            ],
                                            [
                                                Assignment(
                                                    Indexed(
                                                        Variable("Ctx.Cursors"),
                                                        Variable("Fld"),
                                                    ),
                                                    NamedAggregate(
                                                        ("State", Variable("S_Incomplete")),
                                                        ("others", Variable("<>")),
                                                    ),
                                                ),
                                            ],
                                        ),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

    def create_verify_message_procedure(self, message: Message) -> UnitPart:
        specification = ProcedureSpecification(
            "Verify_Message",
            [InOutParameter(["Ctx"], "Context")],
        )

        loop_invariant = And(
            Call("Has_Buffer", [Variable("Ctx")]),
            *common.context_invariant(message, loop_entry=True),
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            Call(
                                self.prefix * message.identifier * "Has_Buffer",
                                [Variable("Ctx")],
                            ),
                        ),
                        Postcondition(
                            And(
                                Call("Has_Buffer", [Variable("Ctx")]),
                                *common.context_invariant(message),
                            ),
                        ),
                    ],
                ),
            ],
            [
                SubprogramBody(
                    specification,
                    [],
                    [
                        ForIn(
                            "F",
                            Variable("Field"),
                            [
                                PragmaStatement("Loop_Invariant", [loop_invariant]),
                                CallStatement("Verify", [Variable("Ctx"), Variable("F")]),
                            ],
                        ),
                    ],
                ),
            ],
        )

    @staticmethod
    def create_present_function() -> UnitPart:
        specification = FunctionSpecification(
            "Present",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    AndThen(
                        Call("Well_Formed", [Indexed(Variable("Ctx.Cursors"), Variable("Fld"))]),
                        Less(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "First"),
                            Add(
                                Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "Last"),
                                Number(1),
                            ),
                        ),
                    ),
                ),
            ],
        )

    @staticmethod
    def create_well_formed_function() -> UnitPart:
        specification = FunctionSpecification(
            "Well_Formed",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Or(
                        *[
                            Equal(
                                Selected(
                                    Indexed(Variable("Ctx.Cursors"), Variable("Fld")),
                                    "State",
                                ),
                                Variable(s),
                            )
                            for s in ("S_Valid", "S_Well_Formed")
                        ],
                    ),
                ),
            ],
        )

    @staticmethod
    def create_valid_function() -> UnitPart:
        specification = FunctionSpecification(
            "Valid",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Postcondition(
                            If(
                                [
                                    (
                                        Result("Valid"),
                                        And(
                                            Call(
                                                "Well_Formed",
                                                [Variable("Ctx"), Variable("Fld")],
                                            ),
                                            Call("Present", [Variable("Ctx"), Variable("Fld")]),
                                        ),
                                    ),
                                ],
                            ),
                        ),
                    ],
                ),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    AndThen(
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Valid"),
                        ),
                        Less(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "First"),
                            Add(
                                Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "Last"),
                                Number(1),
                            ),
                        ),
                    ),
                ),
            ],
        )

    @staticmethod
    def create_incomplete_function() -> UnitPart:
        specification = FunctionSpecification(
            "Incomplete",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Equal(
                        Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                        Variable("S_Incomplete"),
                    ),
                ),
            ],
        )

    @staticmethod
    def create_invalid_function() -> UnitPart:
        specification = FunctionSpecification(
            "Invalid",
            "Boolean",
            [Parameter(["Ctx"], "Context"), Parameter(["Fld"], "Field")],
        )

        return UnitPart(
            [SubprogramDeclaration(specification)],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    Or(
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Invalid"),
                        ),
                        Equal(
                            Selected(Indexed(Variable("Ctx.Cursors"), Variable("Fld")), "State"),
                            Variable("S_Incomplete"),
                        ),
                    ),
                ),
            ],
        )

    def create_well_formed_message_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Well_Formed_Message",
            "Boolean",
            [Parameter(["Ctx"], "Context")],
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            Call(
                                self.prefix * message.identifier * "Has_Buffer",
                                [Variable("Ctx")],
                            ),
                        ),
                    ],
                ),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    self.valid_message_condition(message, well_formed=True),
                ),
            ],
        )

    def create_valid_message_function(self, message: Message) -> UnitPart:
        specification = FunctionSpecification(
            "Valid_Message",
            "Boolean",
            [Parameter(["Ctx"], "Context")],
        )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification,
                    [
                        Precondition(
                            Call(
                                self.prefix * message.identifier * "Has_Buffer",
                                [Variable("Ctx")],
                            ),
                        ),
                    ],
                ),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    self.valid_message_condition(message),
                ),
            ],
        )

    @staticmethod
    def create_incomplete_message_function() -> UnitPart:
        specification = FunctionSpecification(
            "Incomplete_Message",
            "Boolean",
            [Parameter(["Ctx"], "Context")],
        )

        return UnitPart(
            [
                # Eng/RecordFlux/Workarounds#47
                Pragma(
                    "Warnings",
                    [Variable("Off"), String("postcondition does not mention function result")],
                ),
                SubprogramDeclaration(specification, [Postcondition(TRUE)]),
                Pragma(
                    "Warnings",
                    [Variable("On"), String("postcondition does not mention function result")],
                ),
            ],
            private=[
                ExpressionFunctionDeclaration(
                    specification,
                    ForSomeIn(
                        "F",
                        Variable("Field"),
                        Call(
                            "Incomplete",
                            [Variable("Ctx"), Variable("F")],
                        ),
                    ),
                ),
            ],
        )

    def create_scalar_getter_functions(
        self,
        message: Message,
        scalar_fields: Mapping[Field, Scalar],
    ) -> UnitPart:
        def specification(field: Field, field_type: Type) -> FunctionSpecification:
            if field_type.package == BUILTINS_PACKAGE:
                type_identifier = ID(field_type.name)
            else:
                type_identifier = self.prefix * field_type.identifier

            return FunctionSpecification(
                f"Get_{field.name}",
                type_identifier,
                [Parameter(["Ctx"], "Context")],
            )

        def result(field: Field) -> Expr:
            return Call(
                "To_Actual",
                [Selected(Indexed(Variable("Ctx.Cursors"), Variable(field.affixed_name)), "Value")],
            )

        return UnitPart(
            [
                # Eng/RecordFlux/Workarounds#31
                Pragma("Warnings", [Variable("Off"), String("precondition is always False")]),
                *[
                    SubprogramDeclaration(
                        specification(f, t),
                        [
                            Precondition(
                                Call(
                                    self.prefix * message.identifier * "Valid",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                            ),
                        ],
                    )
                    for f, t in scalar_fields.items()
                ],
                Pragma("Warnings", [Variable("On"), String("precondition is always False")]),
            ],
            private=[
                ExpressionFunctionDeclaration(specification(f, t), result(f))
                for f, t in scalar_fields.items()
            ],
        )

    def create_opaque_getter_functions(
        self,
        message: Message,
        opaque_fields: Sequence[Field],
    ) -> UnitPart:
        def name(field: Field) -> str:
            return f"Get_{field.name}"

        def specification(field: Field) -> FunctionSpecification:
            return FunctionSpecification(
                name(field),
                const.TYPES_BYTES,
                [Parameter(["Ctx"], "Context")],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Ghost(),
                        Precondition(
                            AndThen(
                                Call(
                                    self.prefix * message.identifier * "Has_Buffer",
                                    [Variable("Ctx")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Well_Formed",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                            ),
                        ),
                        Postcondition(
                            Equal(
                                Length(Result(name(f))),
                                Call(
                                    const.TYPES_TO_LENGTH,
                                    [
                                        Call(
                                            "Field_Size",
                                            [
                                                Variable("Ctx"),
                                                Variable(f.affixed_name),
                                            ],
                                        ),
                                    ],
                                ),
                            ),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        ReturnStatement(
                            Slice(Variable("Ctx.Buffer.all"), Variable("First"), Variable("Last")),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
        )

    def create_opaque_getter_procedures(
        self,
        message: Message,
        opaque_fields: Sequence[Field],
    ) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Get_{field.name}",
                [Parameter(["Ctx"], "Context"), OutParameter(["Data"], const.TYPES_BYTES)],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            AndThen(
                                Call(
                                    self.prefix * message.identifier * "Has_Buffer",
                                    [Variable("Ctx")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Well_Formed",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Valid_Next",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                                Equal(
                                    Length("Data"),
                                    Call(
                                        const.TYPES_TO_LENGTH,
                                        [
                                            Call(
                                                self.prefix * message.identifier * "Field_Size",
                                                [
                                                    Variable("Ctx"),
                                                    Variable(
                                                        self.prefix
                                                        * message.identifier
                                                        * f.affixed_name,
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                ),
                            ),
                        ),
                        Postcondition(
                            Call(
                                "Equal",
                                [
                                    Variable("Ctx"),
                                    Variable(f.affixed_name),
                                    Variable("Data"),
                                ],
                            ),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        Assignment(
                            "Data",
                            NamedAggregate(("others", First(const.TYPES_BYTE))),
                        ),
                        Assignment(
                            Slice(
                                Variable("Data"),
                                First("Data"),
                                Add(First("Data"), Sub(Variable("Last"), Variable("First"))),
                            ),
                            Slice(
                                Variable("Ctx.Buffer.all"),
                                Variable("First"),
                                Variable("Last"),
                            ),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
        )

    def create_generic_opaque_getter_procedures(
        self,
        message: Message,
        opaque_fields: Sequence[Field],
    ) -> UnitPart:
        def specification(field: Field) -> ProcedureSpecification:
            return ProcedureSpecification(
                f"Generic_Get_{field.name}",
                [Parameter(["Ctx"], "Context")],
            )

        return UnitPart(
            [
                SubprogramDeclaration(
                    specification(f),
                    [
                        Precondition(
                            And(
                                Call(
                                    self.prefix * message.identifier * "Has_Buffer",
                                    [Variable("Ctx")],
                                ),
                                Call(
                                    self.prefix * message.identifier * "Present",
                                    [
                                        Variable("Ctx"),
                                        Variable(self.prefix * message.identifier * f.affixed_name),
                                    ],
                                ),
                            ),
                        ),
                    ],
                    [
                        FormalSubprogramDeclaration(
                            ProcedureSpecification(
                                f"Process_{f.name}",
                                [Parameter([f.name], const.TYPES_BYTES)],
                            ),
                        ),
                    ],
                )
                for f in opaque_fields
            ],
            [
                SubprogramBody(
                    specification(f),
                    [
                        ObjectDeclaration(
                            ["First"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "First",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                        ObjectDeclaration(
                            ["Last"],
                            const.TYPES_INDEX,
                            Call(
                                const.TYPES_TO_INDEX,
                                [
                                    Selected(
                                        Indexed(
                                            Variable("Ctx.Cursors"),
                                            Variable(f.affixed_name),
                                        ),
                                        "Last",
                                    ),
                                ],
                            ),
                            constant=True,
                        ),
                    ],
                    [
                        CallStatement(
                            f"Process_{f.name}",
                            [
                                Slice(
                                    Variable("Ctx.Buffer.all"),
                                    Variable("First"),
                                    Variable("Last"),
                                ),
                            ],
                        ),
                    ],
                )
                for f in opaque_fields
            ],
        )

    def valid_message_condition(self, message: Message, well_formed: bool = False) -> Expr:
        return (
            expr.Or(
                *[
                    expr.AndThen(
                        expr.Call(
                            (
                                "Well_Formed"
                                if well_formed
                                and isinstance(message.field_types[l.source], Composite)
                                else "Valid"
                            ),
                            [
                                expr.Variable("Ctx"),
                                expr.Variable(l.source.affixed_name, immutable=True),
                            ],
                        ),
                        l.condition,
                    )
                    for l in message.incoming(FINAL)
                    if l.target == FINAL
                ],
            )
            .substituted(common.substitution(message, self.prefix))
            .simplified()
            .ada_expr()
        )


def set_context_cursor_scalar() -> Assignment:
    return Assignment(
        Indexed(Variable("Ctx.Cursors"), Variable("Fld")),
        NamedAggregate(
            ("State", Variable("S_Valid")),
            ("First", Call("Field_First", [Variable("Ctx"), Variable("Fld")])),
            ("Last", Call("Field_Last", [Variable("Ctx"), Variable("Fld")])),
            ("Value", Variable("Value")),
        ),
    )


def set_context_cursor_composite_field(field_name: str) -> Assignment:
    return Assignment(
        Indexed(
            Variable("Ctx.Cursors"),
            Variable(field_name),
        ),
        NamedAggregate(
            ("State", Variable("S_Well_Formed")),
            (
                "First",
                Call(
                    "Field_First",
                    [Variable("Ctx"), Variable(field_name)],
                ),
            ),
            (
                "Last",
                Call(
                    "Field_Last",
                    [Variable("Ctx"), Variable(field_name)],
                ),
            ),
            ("Value", Variable("Value")),
        ),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rflx',
 'rflx.converter',
 'rflx.generator',
 'rflx.ide.gnatstudio',
 'rflx.lang',
 'rflx.ls',
 'rflx.model',
 'rflx.pyrflx',
 'rflx.specification']

package_data = \
{'': ['*'], 'rflx': ['ide/vscode/recordflux.vsix', 'templates/*']}

install_requires = \
['attrs>=22.1,<24',
 'defusedxml>=0.7,<0.8',
 'icontract>=2.3.4,<3',
 'importlib_resources>=6,<7',
 'pydantic>=2,<3',
 'pydotplus>=2,<3',
 'pygls>=1.1,<2',
 'ruamel.yaml>=0.18,<0.19',
 'z3-solver>=4,<4.12.3']

entry_points = \
{'console_scripts': ['rflx = rflx.cli:run']}

setup_kwargs = {
    'name': 'RecordFlux',
    'version': '0.20.0',
    'description': 'A toolset for the formal specification and generation of verifiable binary parsers, message generators and protocol state machines.',
    'long_description': "# [RecordFlux](https://github.com/AdaCore/RecordFlux/)\n\n[![PyPI](https://img.shields.io/pypi/v/RecordFlux?color=blue)](https://pypi.org/project/RecordFlux/)\n[![CII Best Practices](https://bestpractices.coreinfrastructure.org/projects/5052/badge)](https://bestpractices.coreinfrastructure.org/projects/5052)\n\nRecordFlux is a toolset for the formal specification and generation of verifiable binary parsers, message generators and protocol state machines.\n\nSee the [website](https://www.adacore.com/recordflux), the [user's guide](https://docs.adacore.com/live/wave/recordflux/html/recordflux_ug/index.html) and the [language reference](https://docs.adacore.com/live/wave/recordflux/html/recordflux_lr/index.html) for more information.\n\n## Contribution and Feedback\n\nContributions and feedback to RecordFlux are very welcome. To discuss a bug or an enhancement, [open a ticket on GitHub](https://github.com/AdaCore/RecordFlux/issues/new/choose) and select the appropriate issue template. Please give sufficient information about your issue, the software version you are using and your environment such that the developers can understand and (if necessary) reproduce the problem. If none of the provided issue templates fit your needs, feel free to open [a blank issue](https://github.com/AdaCore/RecordFlux/issues/new).\n\nSee the [development guide](https://github.com/AdaCore/RecordFlux/blob/main/doc/development_guide/index.rst) on how to contribute to RecordFlux.\n\n## Licence\n\nThis software is licensed under the `AGPL-3.0`. See the `LICENSE` file for the full license text.\n",
    'author': 'Tobias Reiher',
    'author_email': 'reiher@adacore.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://www.adacore.com/recordflux',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.12',
}
from build import *
build(setup_kwargs)

setup(**setup_kwargs)



from __future__ import annotations

import re
from enum import Enum
from pathlib import Path
from typing import Optional

from rflx.error import Location, RecordFluxError, Severity, Subsystem

INCORRECT_LINE_TERMINATORS = "\r"
ILLEGAL_WHITESPACE_CHARACTERS = "\t\x0b\x0c"
KEYWORD_INDENTATION = [
    ("begin", [3, 6]),
    ("end message", [6]),
    ("end", [0, 3, 6]),
    ("for", [3, 12, 15]),
    ("generic", [3]),
    ("goto", [9]),
    ("is", [3, 6]),
    ("message", [6]),
    ("package", [0]),
    ("session", [3]),
    ("state", [6]),
    ("then", [12]),
    ("transition", [6]),
    ("type", [3, 6]),
    ("with", [0, 3, 6, 9, 12, 15]),
]


class Check(Enum):
    ALL = "all"
    BLANK_LINES = "blank-lines"
    CHARACTERS = "characters"
    INDENTATION = "indentation"
    LINE_LENGTH = "line-length"
    TOKEN_SPACING = "token-spacing"
    TRAILING_SPACES = "trailing-spaces"


def check(spec_file: Path) -> RecordFluxError:
    with spec_file.open(encoding="utf-8", newline="") as f:
        return check_string(f.read(), spec_file)


def check_string(specification: str, spec_file: Path = Path("<stdin>")) -> RecordFluxError:
    error = RecordFluxError()

    if specification in ("", "\n"):
        return error

    lines = specification.split("\n")
    enabled_checks = _determine_enabled_checks(error, lines[0], spec_file)

    if not enabled_checks:
        return error

    blank_lines = 0

    for i, l in enumerate(lines, start=1):
        if Check.BLANK_LINES in enabled_checks:
            blank_lines = _check_blank_lines(error, l, i, spec_file, blank_lines, len(lines))
        if Check.CHARACTERS in enabled_checks:
            _check_characters(error, l, i, spec_file)
        if Check.INDENTATION in enabled_checks:
            _check_indentation(error, l, i, spec_file)
        if Check.LINE_LENGTH in enabled_checks:
            _check_line_length(error, l, i, spec_file)
        if Check.TOKEN_SPACING in enabled_checks:
            _check_token_spacing(error, l, i, spec_file)
        if Check.TRAILING_SPACES in enabled_checks:
            _check_trailing_spaces(error, l, i, spec_file)

    return error


def _append(
    error: RecordFluxError,
    message: str,
    row: int,
    col: int,
    spec_file: Path,
    check_type: Optional[Check] = None,
) -> None:
    error.extend(
        [
            (
                message + (f" [{check_type.value}]" if check_type else ""),
                Subsystem.STYLE,
                Severity.ERROR,
                Location((row, col), spec_file),
            ),
        ],
    )


def _determine_enabled_checks(error: RecordFluxError, line: str, spec_file: Path) -> set[Check]:
    checks = {c.value for c in Check.__members__.values()}
    disabled_checks = set()

    m = re.match(r"^\s*--\s*style\s*:\s*disable\s*=\s*([^.]*)$", line)
    if m:
        disabled_checks = {c.strip() for c in m.group(1).split(",")}
        for c in disabled_checks - checks:
            _append(error, f'invalid check "{c}"', 1, 1, spec_file)
    else:
        return set(Check.__members__.values())

    if Check.ALL.value in disabled_checks:
        return set()

    return {Check(c) for c in checks - disabled_checks}


def _check_blank_lines(
    error: RecordFluxError,
    line: str,
    row: int,
    spec_file: Path,
    blank_lines: int,
    row_count: int,
) -> int:
    if not line:
        if row == 1:
            _append(error, "leading blank line", row, 1, spec_file, Check.BLANK_LINES)
        if blank_lines > 0 and row == row_count:
            _append(error, "trailing blank line", row - 1, 1, spec_file, Check.BLANK_LINES)
        blank_lines += 1
    else:
        if blank_lines > 1:
            _append(error, "multiple blank lines", row - 1, 1, spec_file, Check.BLANK_LINES)
        blank_lines = 0

    return blank_lines


def _check_characters(error: RecordFluxError, line: str, row: int, spec_file: Path) -> None:
    for j, c in enumerate(line, start=1):
        if c == INCORRECT_LINE_TERMINATORS:
            s = repr(c).replace("'", '"')
            _append(error, f"incorrect line terminator {s}", row, j, spec_file, Check.CHARACTERS)
        if c in ILLEGAL_WHITESPACE_CHARACTERS:
            s = repr(c).replace("'", '"')
            _append(error, f"illegal whitespace character {s}", row, j, spec_file, Check.CHARACTERS)


def _check_indentation(error: RecordFluxError, line: str, row: int, spec_file: Path) -> None:
    match = re.match(r" *", line)
    assert match
    indentation = match.end()

    for keyword, indentations in KEYWORD_INDENTATION:
        if re.match(rf" *{keyword}", line) and indentation not in indentations:
            expected_indentation = (
                ", ".join(str(row) for row in indentations[:-1]) + f" or {indentations[-1]}"
                if len(indentations) > 1
                else str(indentations[0])
            )
            _append(
                error,
                f"unexpected keyword indentation (expected {expected_indentation})",
                row,
                match.end(),
                spec_file,
                Check.INDENTATION,
            )


def _check_token_spacing(  # noqa: PLR0912
    error: RecordFluxError,
    line: str,
    row: int,
    spec_file: Path,
) -> None:
    strings = [range(m.start(), m.end()) for m in re.finditer(r'"[^"]*"', line.split("--")[0])]

    for match in re.finditer(
        r"--+|/?=(?!>)|<=?|>=?|=>|:=|\+|[(]+\-\d+[)]+|\-|/|\*\*?|::?|'|;|,",
        line,
    ):
        if "--" in line and match.start() > line.find("--"):
            continue

        if any(match.start() in r for r in strings) or any(match.end() in r for r in strings):
            # https://github.com/nedbat/coveragepy/issues/772
            # A dummy statement is needed to disable the peephole optimizer, so that the continue
            # statement is detected during coverage analysis.
            # CPython 3.8 and 3.9 are affected. The issue is fixed in CPython 3.10.
            dummy = 0  # noqa: F841
            continue

        token = match.group(0)

        if token in [";", ","]:
            space_before = False
            space_after = True
        elif token in ["::", "'"]:
            space_before = False
            space_after = False
        else:
            space_before = True
            space_after = True

        if space_before:
            assert token != ";"
            if match.start() > 1 and line[match.start() - 1] not in " (":
                _append(
                    error,
                    f'missing space before "{token}"',
                    row,
                    match.start() + 1,
                    spec_file,
                    Check.TOKEN_SPACING,
                )
        else:
            if match.start() > 1 and line[match.start() - 1] == " ":
                _append(
                    error,
                    f'space before "{token}"',
                    row,
                    match.start() + 1,
                    spec_file,
                    Check.TOKEN_SPACING,
                )
        if space_after:
            if match.end() < len(line) and line[match.end()] not in " ;\n":
                _append(
                    error,
                    f'missing space after "{token}"',
                    row,
                    match.end() + 1,
                    spec_file,
                    Check.TOKEN_SPACING,
                )
        else:
            if match.end() < len(line) and line[match.end()] == " ":
                _append(
                    error,
                    f'space after "{token}"',
                    row,
                    match.end() + 2,
                    spec_file,
                    Check.TOKEN_SPACING,
                )


def _check_trailing_spaces(error: RecordFluxError, line: str, row: int, spec_file: Path) -> None:
    if line.endswith(" "):
        _append(error, "trailing whitespace", row, len(line), spec_file, Check.TRAILING_SPACES)


def _check_line_length(error: RecordFluxError, line: str, row: int, spec_file: Path) -> None:
    if len(line) > 120:
        _append(error, f"line too long ({len(line)}/120)", row, 121, spec_file, Check.LINE_LENGTH)



from __future__ import annotations

import itertools
import logging
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

from rflx import const
from rflx.common import Base, unique, verbose_repr
from rflx.error import RecordFluxError, Severity, Subsystem
from rflx.identifier import ID

from . import message, session, top_level_declaration, type_
from .cache import Cache, Digest
from .package import Package
from .type_ import BUILTIN_TYPES

log = logging.getLogger(__name__)


@dataclass
class UncheckedModel(Base):
    declarations: Sequence[top_level_declaration.UncheckedTopLevelDeclaration]
    error: RecordFluxError

    def checked(
        self,
        cache: Cache,
        workers: int = 1,
    ) -> Model:
        error = RecordFluxError(self.error)
        declarations: list[top_level_declaration.TopLevelDeclaration] = []

        for d in self.declarations:
            try:
                unverified = d.checked(declarations, skip_verification=True)
                digest = Digest(unverified)
                if cache.is_verified(digest):
                    checked = unverified
                else:
                    log.info("Verifying %s", d.identifier)
                    checked = d.checked(declarations, workers=workers)
                declarations.append(checked)
                cache.add_verified(digest)
            except RecordFluxError as e:  # noqa: PERF203
                error.extend(e)

        return Model(declarations, error)


class Model(Base):
    def __init__(
        self,
        declarations: Sequence[top_level_declaration.TopLevelDeclaration] | None = None,
        error: RecordFluxError | None = None,
    ) -> None:
        self._declarations = declarations or []

        error = error or RecordFluxError()
        error += _check_duplicates(self._declarations)
        self._declarations = self._add_type_dependencies(self._declarations)
        error += _check_conflicts(self._declarations)
        error.propagate()

    def __repr__(self) -> str:
        return verbose_repr(self, ["declarations"])

    def __str__(self) -> str:
        return "\n\n".join(self.create_specifications().values())

    @property
    def declarations(self) -> Sequence[top_level_declaration.TopLevelDeclaration]:
        return self._declarations

    @property
    def types(self) -> list[type_.Type]:
        return [d for d in self._declarations if isinstance(d, type_.Type)]

    @property
    def messages(self) -> list[message.Message]:
        return [d for d in self._declarations if isinstance(d, message.Message)]

    @property
    def refinements(self) -> list[message.Refinement]:
        return [d for d in self._declarations if isinstance(d, message.Refinement)]

    @property
    def sessions(self) -> list[session.Session]:
        return [d for d in self._declarations if isinstance(d, session.Session)]

    @property
    def packages(self) -> dict[ID, list[top_level_declaration.TopLevelDeclaration]]:
        return {p: list(d) for p, d in itertools.groupby(self._declarations, lambda x: x.package)}

    def create_specifications(self) -> dict[ID, str]:
        pkgs: dict[ID, Package] = {}
        for d in self.declarations:
            if isinstance(d, type_.Type):
                if not type_.is_builtin_type(d.name) and not type_.is_internal_type(d.name):
                    pkg_name: ID = d.package
                    pkg = pkgs.setdefault(pkg_name, Package(pkg_name))
                    pkg.imports |= {dep.package for dep in d.direct_dependencies}
                    pkg.declarations.append(d)
            else:
                pkg_name = d.package
                pkgs.setdefault(pkg_name, Package(pkg_name)).declarations.append(d)
        return {identifier: str(pkg) for identifier, pkg in pkgs.items()}

    def write_specification_files(self, output_dir: Path) -> None:
        """Write corresponding specification files (one per package) into given directory."""
        for package, specification in self.create_specifications().items():
            header = (
                "-- style: disable = line-length\n\n"
                if any(len(l) > 120 for l in specification.split("\n"))
                else ""
            )
            (output_dir / f"{package.flat.lower()}.rflx").write_text(f"{header}{specification}")

    @staticmethod
    def _add_type_dependencies(
        declarations: Sequence[top_level_declaration.TopLevelDeclaration],
    ) -> list[top_level_declaration.TopLevelDeclaration]:
        """Add missing type dependencies to list of declarations."""
        result: list[top_level_declaration.TopLevelDeclaration] = []

        for d in declarations:
            if isinstance(d, type_.Type):
                result.extend(d.dependencies)
                result.append(d)

            if isinstance(d, session.Session):
                result.extend(d.direct_dependencies.values())
                result.append(d)

        return list(unique(result))


def _check_duplicates(
    declarations: Sequence[top_level_declaration.TopLevelDeclaration],
) -> RecordFluxError:
    error = RecordFluxError()
    seen: dict[ID, top_level_declaration.TopLevelDeclaration] = {}

    for d in declarations:
        if d.package in (const.BUILTINS_PACKAGE, const.INTERNAL_PACKAGE):
            continue

        if type_.is_builtin_type(d.identifier.name) or type_.is_internal_type(d.identifier.name):
            error.extend(
                [
                    (
                        f'illegal redefinition of built-in type "{d.identifier.name}"',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        d.location,
                    ),
                ],
            )

        elif d.identifier in seen:
            error.extend(
                [
                    (
                        (
                            f'conflicting refinement of "{d.pdu.identifier}" with'
                            f' "{d.sdu.identifier}"'
                            if isinstance(d, message.Refinement)
                            else (
                                f'name conflict for type "{d.identifier}"'
                                if isinstance(d, type_.Type)
                                else f'name conflict for session "{d.identifier}"'
                            )
                        ),
                        Subsystem.MODEL,
                        Severity.ERROR,
                        d.location,
                    ),
                    (
                        (
                            "previous occurrence of refinement"
                            if isinstance(d, message.Refinement)
                            else f'previous occurrence of "{d.identifier}"'
                        ),
                        Subsystem.MODEL,
                        Severity.INFO,
                        seen[d.identifier].location,
                    ),
                ],
            )
        seen[d.identifier] = d

    return error


def _check_conflicts(
    declarations: Sequence[top_level_declaration.TopLevelDeclaration],
) -> RecordFluxError:
    error = RecordFluxError()

    for e1, e2 in [
        (e1, e2)
        for i1, e1 in enumerate(declarations)
        for i2, e2 in enumerate(declarations)
        if (
            isinstance(e1, type_.Enumeration)
            and isinstance(e2, type_.Enumeration)
            and i1 < i2
            and (
                e1.package in (e2.package, const.BUILTINS_PACKAGE)
                or e2.package == const.BUILTINS_PACKAGE
            )
        )
    ]:
        identical_literals = set(e2.literals) & set(e1.literals)

        if identical_literals:
            literals_message = ", ".join([f"{l}" for l in sorted(identical_literals)])
            error.extend(
                [
                    (
                        f"conflicting literals: {literals_message}",
                        Subsystem.MODEL,
                        Severity.ERROR,
                        e2.location,
                    ),
                    *[
                        (
                            f'previous occurrence of "{l}"',
                            Subsystem.MODEL,
                            Severity.INFO,
                            l.location,
                        )
                        for l in sorted(identical_literals)
                    ],
                ],
            )

    literals = [
        ID(d.package * l, location=l.location)
        for d in declarations
        if isinstance(d, type_.Enumeration)
        for l in d.literals
    ]
    name_conflicts = {
        *(set(literals) & {d.identifier for d in declarations}),
        *({l.name for l in literals} & {t.name for t in BUILTIN_TYPES}),
    }
    if name_conflicts:
        for literal, conflicting_type in [
            (l, d)
            for l in literals
            for d in declarations
            if {l, l.name} & name_conflicts and l.name == d.identifier.name
        ]:
            error.extend(
                [
                    (
                        f'literal "{literal.name}" conflicts with type declaration',
                        Subsystem.MODEL,
                        Severity.ERROR,
                        literal.location,
                    ),
                    (
                        f'conflicting type "{conflicting_type.identifier}"',
                        Subsystem.MODEL,
                        Severity.INFO,
                        conflicting_type.location,
                    ),
                ],
            )

    return error



from __future__ import annotations

import typing as ty
from pathlib import Path
from typing import Optional

from annotated_types import Gt
from pydantic import BaseModel, ConfigDict, Field, ValidationError
from ruamel.yaml.error import MarkedYAMLError
from ruamel.yaml.main import YAML
from typing_extensions import Annotated

from rflx.error import Location, RecordFluxError, Severity, Subsystem
from rflx.identifier import ID
from rflx.model import Model
from rflx.model.session import Session

# TODO(eng/recordflux/RecordFlux#1359): Replace ty.* by collections.abc.*
# Sequence and Mapping are imported from collections.abc as importing them
# from typing is deprecated. However pydantic does not support the imported
# version from collections.abc. To fix that typing is imported as ty and the
# typing versions of Sequence and Mapping are used in classes that derive
# from pydantic.BaseModel.
# This is only relevant for Python 3.8.


IntSize = Annotated[int, Gt(0)]


class SessionSize(BaseModel):  # type: ignore[misc]
    default: Optional[IntSize] = Field(None, alias="Default")
    global_: Optional[ty.Mapping[str, IntSize]] = Field(None, alias="Global")
    local_: Optional[ty.Mapping[str, ty.Mapping[str, IntSize]]] = Field(None, alias="Local")

    model_config = ConfigDict(extra="forbid")


class SessionIntegration(BaseModel):  # type: ignore[misc]
    buffer_size: SessionSize = Field(alias="Buffer_Size")

    model_config = ConfigDict(extra="forbid")


class IntegrationFile(BaseModel):  # type: ignore[misc]
    session: ty.Mapping[str, SessionIntegration] = Field(alias="Session")

    model_config = ConfigDict(extra="forbid")


class Integration:
    @property
    def defaultsize(self) -> int:
        return 4096

    def __init__(self, integration_files_dir: Optional[Path] = None) -> None:
        self._packages: dict[str, IntegrationFile] = {}
        self._integration_files_dir = integration_files_dir

    def load_integration_file(self, spec_file: Path, error: RecordFluxError) -> None:
        integration_file = (
            spec_file.with_suffix(".rfi")
            if self._integration_files_dir is None
            else self._integration_files_dir / (spec_file.stem + ".rfi")
        )
        if integration_file.exists():
            yaml = YAML()
            try:
                content = yaml.load(integration_file)
            except MarkedYAMLError as e:
                location = Location(
                    start=(
                        (0, 0)
                        if e.problem_mark is None
                        else (e.problem_mark.line + 1, e.problem_mark.column + 1)
                    ),
                    source=integration_file,
                )
                error.extend([(str(e), Subsystem.PARSER, Severity.ERROR, location)])
                return
            self._add_integration_object(integration_file, content, error)

    def validate(self, model: Model, error: RecordFluxError) -> None:
        for package, integration_file in self._packages.items():
            for session_name, integration in integration_file.session.items():
                matching_sessions = [
                    s
                    for s in model.sessions
                    if package == str(s.package).lower() and str(s.identifier.name) == session_name
                ]
                if not matching_sessions:
                    error.extend(
                        [
                            (
                                f'unknown session "{session_name}"',
                                Subsystem.PARSER,
                                Severity.ERROR,
                                Integration._to_location(package),
                            ),
                        ],
                    )
                    return
                assert len(matching_sessions) == 1
                session = matching_sessions[0]
                self._validate_globals(package, integration, session, error)
                self._validate_states(package, integration, session, error)

    def get_size(self, session: ID, variable: Optional[ID], state: Optional[ID]) -> int:
        """
        Return the requested buffer size for a variable of a given session and state.

        If state is None, the variable is assumed to be a global variable. If variable is None or no
        specific buffer size was requested for the variable, return the default buffer size for the
        session, if present, or the default buffer size for RecordFlux.

        The returned size is in bytes.
        """
        integration_package = str(session.parent).lower()
        if integration_package not in self._packages:
            return self.defaultsize

        session_name = str(session.name)
        if session_name not in self._packages[integration_package].session:
            return self.defaultsize

        buffer_size = self._packages[integration_package].session[session_name].buffer_size
        default_size = self.defaultsize if buffer_size.default is None else buffer_size.default

        if variable is None:
            return default_size

        variable_name = str(variable.name)
        if state is None:
            if buffer_size.global_ is not None and variable_name in buffer_size.global_:
                return buffer_size.global_[variable_name]
            return default_size

        state_name = str(state)
        if (
            buffer_size.local_ is not None
            and state_name in buffer_size.local_
            and variable_name in buffer_size.local_[state_name]
        ):
            return buffer_size.local_[state_name][variable_name]

        if buffer_size.global_ is not None and variable_name in buffer_size.global_:
            return buffer_size.global_[variable_name]

        return default_size

    def _add_integration_object(self, filename: Path, file: object, error: RecordFluxError) -> None:
        try:
            self._packages[filename.stem] = IntegrationFile.model_validate(file)
        except ValidationError as e:
            error.extend(
                [(f"{e}", Subsystem.PARSER, Severity.ERROR, self._to_location(filename.stem))],
            )

    @staticmethod
    def _to_location(package: str) -> Location:
        return Location(start=(0, 0), source=Path(package + ".rfi"))

    @staticmethod
    def _validate_globals(
        package: str,
        integration: SessionIntegration,
        session: Session,
        error: RecordFluxError,
    ) -> None:
        if integration.buffer_size.global_ is None:
            return
        session_decl_vars = [str(x.name) for x in session.declarations]
        for var_name in integration.buffer_size.global_:
            if var_name not in session_decl_vars:
                error.extend(
                    [
                        (
                            (
                                f'unknown global variable "{var_name}" '
                                f'in session "{session.identifier.name}"'
                            ),
                            Subsystem.PARSER,
                            Severity.ERROR,
                            Integration._to_location(package),
                        ),
                    ],
                )

    @staticmethod
    def _validate_states(
        package: str,
        integration: SessionIntegration,
        session: Session,
        error: RecordFluxError,
    ) -> None:
        if integration.buffer_size.local_ is None:
            return
        for state_name, state_entry in integration.buffer_size.local_.items():
            state = None
            for s in session.states:
                if str(s.identifier.name) == state_name:
                    state = s
            if state is None:
                error.extend(
                    [
                        (
                            (
                                f'unknown state "{state_name}" in session '
                                f'"{session.identifier.name}"'
                            ),
                            Subsystem.PARSER,
                            Severity.ERROR,
                            Integration._to_location(package),
                        ),
                    ],
                )
                return
            state_declaration_vars = [str(x.name) for x in state.declarations]
            for var_name in state_entry:
                if var_name not in state_declaration_vars:
                    error.extend(
                        [
                            (
                                (
                                    f'unknown variable "{var_name}" in state '
                                    f'"{state_name}" of session "{session.identifier.name}"'
                                ),
                                Subsystem.PARSER,
                                Severity.ERROR,
                                Integration._to_location(package),
                            ),
                        ],
                    )



from __future__ import annotations

import os
import re
import textwrap
from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path
from typing import NoReturn, TypeVar


def format_repr(string: str) -> str:
    """
    Improve formatting of string representation.

    When `RFLX_TESTING` is set, Black is used to format the string representation. This makes
    pytest diffs of complex objects more readable, especially when using
    [pytest-clarity](https://github.com/darrenburns/pytest-clarity) or
    [pytest-icdiff](https://github.com/hjwp/pytest-icdiff).
    """
    if os.environ.get("RFLX_TESTING"):
        from black import FileMode, format_str

        return format_str(string, mode=FileMode(line_length=60))

    return string


STDIN = Path("<stdin>")


class Base:
    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            for k in other.__dict__:
                if k not in ["location", "error"] and k not in self.__dict__:
                    return False
            for k, v in self.__dict__.items():
                if k not in ["location", "error"] and (
                    k not in other.__dict__ or v != other.__dict__[k]
                ):
                    return False
            return True
        return NotImplemented

    def __repr__(self) -> str:
        args = "\n" + ",\n".join(f"{k}={v!r}" for k, v in self.__dict__.items() if k != "location")
        return format_repr(indent_next(f"\n{self.__class__.__name__}({indent(args, 4)})", 4))


def verbose_repr(obj: object, attributes: Sequence[str]) -> str:
    def prefixed_str(obj: object) -> str:
        obj_str = str(obj)
        return (
            "\n" + "\n".join(f"# {l}" if l else "#" for l in obj_str.split("\n")) + "\n"
            if obj_str
            else ""
        )

    indentation = len(obj.__class__.__name__) + 1
    args = "".join(f"{a}={getattr(obj, a)!r},\n" for a in attributes)
    return format_repr(
        indent_next(
            f"\n{obj.__class__.__name__}({indent_next(args, indentation)}\n){prefixed_str(obj)}",
            4,
        ),
    )


def indent(string: str, indentation: int) -> str:
    return textwrap.indent(string, indentation * " ")


def indent_next(string: str, indentation: int) -> str:
    return string.replace("\n", "\n" + indentation * " ")


def file_name(name: str) -> str:
    return re.sub(r"(?:\.|::)", "-", name.lower())


T = TypeVar("T")


def unique(iterable: Iterable[T]) -> Iterator[T]:
    # In Python 3.7+, standard `dict` is guaranteed to preserve order:
    # https://stackoverflow.com/a/39980744
    return iter(dict.fromkeys(iterable))


def assert_never(value: NoReturn) -> NoReturn:
    assert False, f'unhandled value "{value}" ({type(value).__name__})'



from __future__ import annotations

import re
import string
import textwrap
from abc import ABC, abstractmethod
from collections.abc import Sequence
from datetime import datetime
from pathlib import Path
from typing import Optional, Union
from xml.etree.ElementTree import Element, ParseError

from defusedxml import ElementTree

import rflx.specification.const
from rflx.common import file_name
from rflx.error import Location, RecordFluxError, Subsystem, fail, warn

NAMESPACE = {"iana": "http://www.iana.org/assignments"}
RESERVED_WORDS = "|".join(rflx.specification.const.RESERVED_WORDS)
OUTPUT_INDENT_CHAR = " "
OUTPUT_INDENT = OUTPUT_INDENT_CHAR * 3
OUTPUT_WIDTH = 80


def convert(
    data: str,
    source: Path,
    always_valid: bool,
    output_dir: Path,
    reproducible: bool = False,
) -> None:
    try:
        root = ElementTree.fromstring(data)
    except ParseError as e:
        fail(
            f"invalid XML document: {e}",
            subsystem=Subsystem.CONVERTER,
            location=Location(start=e.position, source=source),
        )
    registry_id = root.get("id")
    if registry_id is None:
        fail(
            "no registry ID found",
            subsystem=Subsystem.CONVERTER,
            location=Location(start=(0, 0), source=source),
        )
    package_name = _normalize_name(registry_id)
    registry_last_updated = root.find("iana:updated", NAMESPACE)
    registry_title = root.find("iana:title", NAMESPACE)
    package_entries: list[SpecificationElement] = []
    for sub_registry in root.findall(root.tag):
        try:
            enum_type = _convert_registry_to_enum_type(sub_registry, always_valid)
            if enum_type is not None:
                package_entries.append(enum_type)
        except RecordFluxError as e:  # noqa: PERF203
            sub_registry_id = sub_registry.get("id")
            sub_registry_title = sub_registry.find("iana:title", NAMESPACE)
            title_str = f'"{sub_registry_title.text}"' if sub_registry_title is not None else ""
            id_str = f' (id={sub_registry_id if sub_registry_id is not None else "<Unknown>"})'
            err_msg = f"{e}"[len("converter: error: ") :]
            warn(
                f"registry {title_str}{id_str} skipped due to conversion error: {err_msg}",
                subsystem=Subsystem.CONVERTER,
            )
            package_entries.append(
                CommentBlock(
                    f"type {_normalize_name(title_str)}\n"
                    "... Skipped due to unsupported content in the registry ...",
                    trailing_lines=1,
                ),
            )
    enum_types = [x for x in package_entries if isinstance(x, EnumType)]
    resolve_duplicate_literals(enum_types)
    file = output_dir / Path(file_name(package_name) + ".rflx")
    write_rflx_specification(
        file,
        package_entries,
        package_name,
        registry_title,
        registry_last_updated,
        reproducible,
    )


def write_rflx_specification(
    file: Path,
    package_entries: list[SpecificationElement],
    package_name: str,
    registry_title: Optional[Element],
    registry_last_updated: Optional[Element],
    reproducible: bool,
) -> None:
    package_header: list[CommentBlock] = []

    package_header.append(
        CommentBlock(
            "AUTOMATICALLY GENERATED. DO NOT EDIT."
            + (
                ""
                if reproducible
                else f"\nGeneration date: {datetime.now().strftime('%Y-%m-%d')}"  # noqa: DTZ005
            ),
        ),
    )
    if registry_title is not None:
        package_header.append(CommentBlock(f"{registry_title.text}"))
    if registry_last_updated is not None:
        package_header.append(
            CommentBlock(
                f"Registry last updated on {registry_last_updated.text}",
                trailing_lines=1,
            ),
        )

    package = Package(package_name, package_header, package_entries)
    file.write_text(package.rflx_str(), encoding="utf-8")


def resolve_duplicate_literals(enum_types: Sequence[EnumType]) -> None:
    literal_names = [
        enum_literal.name for enum_type in enum_types for enum_literal in enum_type.enum_literals
    ]
    for enum_type in enum_types:
        for enum_literal in enum_type.enum_literals:
            if literal_names.count(enum_literal.name) > 1:
                enum_literal.name += f"_{enum_literal.value.replace('16#', '').replace('#', '')}"


def _convert_registry_to_enum_type(
    registry: Element,
    always_valid: bool,
) -> Optional[EnumType]:
    records = registry.findall("iana:record", NAMESPACE)
    title = registry.find("iana:title", NAMESPACE)
    if len(records) == 0 or title is None or title.text is None:
        return None

    registry_title = _normalize_name(title.text)
    enum_literals: dict[str, EnumLiteral] = {}
    enum_literal_highest_bit_len = 0
    for record in records:
        name_tag = _get_name_tag(record)
        value_tag = "value"

        if name_tag is None:
            continue

        name_element = record.find(f"iana:{name_tag}", NAMESPACE)
        value_element = record.find(f"iana:{value_tag}", NAMESPACE)

        if value_element is None or name_element is None:
            # https://github.com/nedbat/coveragepy/issues/772
            # A dummy statement is needed to disable the peephole optimizer, so that the continue
            # statement is detected during coverage analysis.
            # CPython 3.8 and 3.9 are affected. The issue is fixed in CPython 3.10.
            dummy = 0  # noqa: F841
            continue

        assert isinstance(value_element.text, str)
        assert isinstance(name_element.text, str)

        rflx_value, bit_length = _normalize_value(value_element.text)

        rflx_name = _normalize_literal_name(name_element.text)

        if bit_length > enum_literal_highest_bit_len:
            enum_literal_highest_bit_len = bit_length

        if re.search(r"RESERVED|UNASSIGNED", rflx_name, flags=re.I):
            continue

        if re.fullmatch(RESERVED_WORDS, rflx_name, re.I | re.X) is not None:
            rflx_name += f"_{rflx_value.replace('16#', '').replace('#', '')}"
        if rflx_name.startswith(tuple(d for d in string.digits)):
            rflx_name = f"{registry_title}_{rflx_name}"

        comment = [
            element
            for element in record.iterfind("*", NAMESPACE)
            if element.tag
            not in [f"{{{NAMESPACE['iana']}}}{name_tag}", f"{{{NAMESPACE['iana']}}}{value_tag}"]
        ]

        enum_literal = EnumLiteral(rflx_name, rflx_value, bit_length, comment)
        if rflx_value in enum_literals:
            enum_literals[rflx_value].join(enum_literal, registry_title)
        else:
            enum_literals[rflx_value] = enum_literal

    if len(enum_literals.values()) > 0:
        return EnumType(
            registry_title,
            list(enum_literals.values()),
            enum_literal_highest_bit_len,
            always_valid,
        )
    return None


def _get_name_tag(record: Element) -> Optional[str]:
    sub_elements = record.findall("*", NAMESPACE)
    child_names = {c.tag[c.tag.index("}") + 1 :] for c in sub_elements}
    possible_name_tags = ["name", "code", "type", "description"]
    for tag in possible_name_tags:
        if tag in child_names:
            return tag
    return None


class SpecificationElement(ABC):
    """Base class for RecordFlux specification elements."""

    def __str__(self) -> str:
        """
        Render the current element in RecordFlux concrete syntax.

        Note: If relative indentation is needed, then the method 'rflx_str' should be used instead.
        """
        return self.rflx_str()

    @abstractmethod
    def rflx_str(self, indent_level: int = 0) -> str:
        """Render the current element in RecordFlux concrete syntax with given base indentation."""


class Package(SpecificationElement):
    def __init__(
        self,
        name: str,
        header: Sequence[CommentBlock],
        entries: Sequence[SpecificationElement],
    ):
        self.name = name
        self.header = header
        self.entries = entries

    def rflx_str(self, indent_level: int = 0) -> str:
        base_indent = OUTPUT_INDENT * indent_level
        str_repr = ""
        str_repr += "".join(r.rflx_str(indent_level) for r in self.header)
        str_repr += f"{base_indent}package {self.name} is\n\n"
        str_repr += "".join(r.rflx_str(indent_level + 1) for r in self.entries)
        str_repr += f"{base_indent}end {self.name};\n"
        return str_repr


class EnumType(SpecificationElement):
    def __init__(
        self,
        type_name: str,
        enum_literals: Sequence[EnumLiteral],
        type_size: int,
        always_valid: bool,
    ) -> None:
        self.type_name = type_name
        self.enum_literals = enum_literals
        self.type_size = type_size
        self.always_valid = always_valid and len(self.enum_literals) < 2**self.type_size

    def rflx_str(self, indent_level: int = 0) -> str:
        base_indent = OUTPUT_INDENT * indent_level
        formatted_enum_literals = ",\n".join(
            r.rflx_str(indent_level + 1) for r in self.enum_literals
        )
        return (
            f"{base_indent}type {self.type_name} is\n"
            f"{base_indent + OUTPUT_INDENT}({formatted_enum_literals.lstrip()})\n"
            f"{base_indent}with Size => {self.type_size}"
            f"{', Always_Valid;' if self.always_valid else ';'}"
            f"\n\n"
        )


class EnumLiteral(SpecificationElement):
    def __init__(
        self,
        rflx_name: str,
        rflx_value: str,
        bit_length: int,
        comments: Optional[list[Element]] = None,
    ):
        self.name = rflx_name
        self.value = rflx_value
        self.bit_length = bit_length
        self.comment_list = comments or []
        self.alternative_names: list[str] = [f"alternative_name = {self.name}"]

    def join(self, duplicate: EnumLiteral, registry_name: str) -> None:
        self.name = f"{registry_name}_{self.value.replace('16#', '').replace('#', '')}"
        self.comment_list.extend(duplicate.comment_list)
        self.alternative_names.append(f"alternative_name = {duplicate.name}")

    def comment(self, indent_level: int) -> list[str]:
        base_indent = OUTPUT_INDENT * indent_level
        comments = [
            (
                f"{c.tag[c.tag.index('}') + 1:]} = {c.text}"
                if c.tag is not None and c.text is not None
                else f"Ref: {c.attrib['data']}"
            )
            for c in self.comment_list
        ]
        if len(self.alternative_names) > 1:
            comments.extend(self.alternative_names)

        formatted_comments = [
            textwrap.wrap(comment, width=OUTPUT_WIDTH, subsequent_indent=f"{base_indent} --  ")
            for comment in comments
        ]
        if len(formatted_comments) > 0:
            return ["\n".join(formatted_comment) for formatted_comment in formatted_comments]
        return []

    def rflx_str(self, indent_level: int = 0) -> str:
        base_indent = OUTPUT_INDENT * indent_level
        str_repr = ""
        comment = self.comment(indent_level)
        if comment:
            str_repr += "\n"
            for comment_line in comment:
                str_repr += f"{base_indent} -- {comment_line}\n"
        name = (
            self.name
            if len(self.name) <= 100
            else self.name[:100]
            + ("_" if self.name[99] != "_" else "")
            + self.value.replace("16#", "").replace("#", "")
        )
        str_repr += f"{base_indent} {f'{name} => {self.value}'}"
        return str_repr


class CommentBlock(SpecificationElement):
    """
    A textual comment block in the generated source code.

    An instance of this class holds one text string that can contain multiple lines. Comment
    characters (i.e., '--' for RecordFlux) and leading whitespace should *not* be included. When the
    object is converted to string leading whitespace and comment characters are added automatically.
    Lines that would be longer than 80 characters are wrapped. However, existing line-endings are
    preserved.
    """

    def __init__(
        self,
        text: str,
        trailing_lines: int = 0,
    ) -> None:
        self.text = text
        self.trailing_lines = trailing_lines

    def rflx_str(self, indent_level: int = 0) -> str:
        def wrap_line(line: str) -> str:
            commented_section = textwrap.wrap(
                f"{OUTPUT_INDENT*indent_level}-- {line}",
                width=OUTPUT_WIDTH,
                subsequent_indent=f"{OUTPUT_INDENT*indent_level}-- ",
            )
            return "\n".join(commented_section)

        lines = self.text.split("\n")
        wrapped_lines = map(wrap_line, lines)
        return "\n".join(wrapped_lines) + "\n" + "\n" * self.trailing_lines


def _normalize_name(description_text: str) -> str:
    t: dict[str, Union[int, str, None]] = {c: " " for c in string.punctuation + "\n"}
    name = description_text.translate(str.maketrans(t))
    return "_".join([s[0].upper() + s[1:] for s in name.split()])


def _normalize_literal_name(description_text: str) -> str:
    """
    Normalize the given text to be used as the name of an enum literal.

    The input text is expected to be a potentially multiline string that
    sometimes contains additional comments in parentheses. Only the part up to
    the first parenthesis is kept. The rest is normalized in the same way
    as in _normalize_name. In addition, if the original string contains either
    "(deprecated)" or "(obsolete)", then those words are prefixed to the
    normalized name.
    """
    description_prefix = re.sub(r"\(.*", "", description_text, flags=re.S)
    rflx_name = _normalize_name(description_prefix)
    if re.search(r"\(obsolete\)", description_text, re.IGNORECASE | re.S):
        rflx_name = "Obsolete_" + rflx_name
    if re.search(r"\(deprecated\)", description_text, re.IGNORECASE | re.S):
        rflx_name = "Deprecated_" + rflx_name
    return rflx_name


def _normalize_value(value: str) -> tuple[str, int]:
    """
    Return hex string and required bit length.

    In case of a range, just return the upper bound of that range. Ranges can only occur for
    UNASSIGNED or RESERVED values, which are filtered out.
    """
    if value.find("0x") != -1:
        rflx_hex = _normalize_hex_value(value)
        return rflx_hex, len(rflx_hex[3 : len(rflx_hex) - 1].lstrip("0")) * 4
    if value.find("-") != -1:
        range_upper = value[value.index("-") + 1 :]
        return range_upper, _convert_int_value(range_upper).bit_length()
    return str(value), _convert_int_value(value).bit_length()


def _normalize_hex_value(hex_value: str) -> str:
    if hex_value.find("-") != -1 or re.match(r"^0x[0-9A-Fa-f]{2},\*", hex_value) is not None:
        if re.match(r"^0x[0-9A-Fa-f]{4}-0x[0-9A-Fa-f]{4}$", hex_value) is not None:  # 0x081D-0x08FF
            return f"16#{hex_value[9:]}#".upper()
        if re.match(r"^0x[0-9A-Fa-f]{2}-[0-9A-Fa-f]{2},\*$", hex_value) is not None:  # 0xCD-CF,*
            return f"16#{hex_value[5:7]}FF#".upper()
        if re.match(r"^0x[0-9A-F]{2},\*", hex_value) is not None:  # 0xCB,*
            return f"16#{hex_value[2:4]}FF#".upper()
        if (
            re.match(r"^0x[0-9A-Fa-f]{2},0x[0-9A-Fa-f]{2}-[0-9A-Fa-f]{2}$", hex_value) is not None
        ):  # 0xCC,0xAF-FF
            return f"16#{hex_value[2:4]}{hex_value[10:12]}#".upper()
        fail(f'cannot normalize hex value range "{hex_value}"', subsystem=Subsystem.CONVERTER)
    if re.match(r"^0x[0-9A-Fa-f]{2},0x[0-9A-Fa-f]{2}$", hex_value) is not None:  # 0x0A,0xFF
        return f"16#{hex_value.replace('0x', '').replace(',', '')}#".upper()
    if re.match(r"^0x[0-9A-Fa-f]+$", hex_value) is not None:  # 0xA1A1
        return f"16#{hex_value[2:]}#".upper()
    fail(f'cannot normalize hex value "{hex_value}"', subsystem=Subsystem.CONVERTER)


def _convert_int_value(value: str) -> int:
    try:
        return int(value)
    except ValueError as e:
        fail(f"{e}", subsystem=Subsystem.CONVERTER)



"""Intermediate representation."""

from __future__ import annotations

import re
from abc import abstractmethod
from collections.abc import Generator, Mapping, Sequence
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from sys import intern
from typing import TYPE_CHECKING, Optional, Protocol, TypeVar, Union

import z3
from attr import define, field, frozen

from rflx import typing_ as rty
from rflx.common import Base
from rflx.const import MAX_SCALAR_SIZE, MP_CONTEXT
from rflx.error import Location, Subsystem, info
from rflx.identifier import ID, StrID

if TYPE_CHECKING:
    from rflx.model import type_ as mty

INT_MIN: int = 0
INT_MAX: int = 2**MAX_SCALAR_SIZE - 1


class Origin(Protocol):
    def __str__(self) -> str: ...  # pragma: no cover

    @property
    def location(self) -> Optional[Location]: ...  # pragma: no cover


class ConstructedOrigin(Origin):
    def __init__(self, string_representation: str, location: Optional[Location]) -> None:
        self._string_representation = string_representation
        self._location = location

    def __str__(self) -> str:
        return self._string_representation

    @property
    def location(self) -> Optional[Location]:
        return self._location


class ProofResult(Enum):
    SAT = z3.sat
    UNSAT = z3.unsat
    UNKNOWN = z3.unknown


class ProofJob(Base):
    def __init__(
        self,
        facts: Sequence[Stmt],
        results: Mapping[ProofResult, Sequence[int]],
        logic: str = "QF_NIA",
    ):
        self._facts = facts
        self._logic = logic
        self._results = results
        self._result: Sequence[int]

    @property
    def result(self) -> Sequence[int]:
        return self._result

    def check(self) -> ProofJob:
        """Check the specified facts and return the corresponding object depending on the result."""

        solver = z3.SolverFor(self._logic)

        for f in self._facts:
            solver.add(f.to_z3_expr())

        proof_result = ProofResult(solver.check())
        self._result = self._results[proof_result]

        return self


class ProofManager(Base):
    def __init__(self, workers: int) -> None:
        self._jobs: list[ProofJob] = []
        self._workers = workers

    def add(self, jobs: Sequence[ProofJob]) -> None:
        self._jobs.extend(jobs)

    def check(self) -> list[ProofJob]:
        with ProcessPoolExecutor(max_workers=self._workers, mp_context=MP_CONTEXT) as executor:
            result = list(executor.map(ProofManager._check, self._jobs))

        self._jobs.clear()

        return result

    @staticmethod
    def _check(job: ProofJob) -> ProofJob:
        return job.check()


class Cond(Base):
    def __init__(self, goal: BoolExpr, facts: Optional[Sequence[Stmt]] = None) -> None:
        self._goal = goal
        self._facts = facts or []

    @property
    def goal(self) -> BoolExpr:
        return self._goal

    @property
    def facts(self) -> Sequence[Stmt]:
        return self._facts


@define(init=False, eq=False)
class Stmt(Base):
    """Statement in three-address code (TAC) format."""

    origin: Optional[Origin]
    _str: Optional[str] = field(init=False, default=None)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return str(self) == str(other)
        return NotImplemented

    def __str__(self) -> str:
        if self._str is not None:
            return self._str

        self._update_str()
        assert self._str is not None
        return self._str  # type: ignore[unreachable]

    @property
    def location(self) -> Optional[Location]:
        return self.origin.location if self.origin else None

    def substituted(self, mapping: Mapping[ID, ID]) -> Stmt:
        raise NotImplementedError

    @abstractmethod
    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        raise NotImplementedError

    @abstractmethod
    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def precondition_proofs(self, facts: Sequence[Stmt]) -> list[ProofJob]:
        raise NotImplementedError

    @abstractmethod
    def _update_str(self) -> None:
        raise NotImplementedError


@define(eq=False)
class VarDecl(Stmt):
    identifier: ID = field(converter=ID)
    type_: rty.NamedType
    expression: Optional[ComplexExpr] = None
    origin: Optional[Origin] = None

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.BoolRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        initialization = f" = {self.expression}" if self.expression else ""
        self._str = intern(f"Var {self.identifier} : {self.type_.identifier}{initialization}")


@define(eq=False)
class Assign(Stmt):
    target: ID = field(converter=ID)
    expression: Expr
    type_: rty.NamedType
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> Assign:
        return Assign(
            mapping.get(self.target, self.target),
            self.expression.substituted(mapping),
            self.type_,
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        target: Var
        if isinstance(self.expression, IntExpr):
            target = IntVar(self.target, self.expression.type_)
        elif isinstance(self.expression, BoolExpr):
            target = BoolVar(self.target)
        else:
            return z3.BoolVal(val=True)
        return target.to_z3_expr() == self.expression.to_z3_expr()

    @property
    def target_var(self) -> Var:
        if isinstance(self.expression, IntExpr):
            return IntVar(self.target, self.expression.type_)
        if isinstance(self.expression, BoolExpr):
            return BoolVar(self.target)
        assert isinstance(self.expression, ObjVar)
        return ObjVar(self.target, self.expression.type_)

    def _update_str(self) -> None:
        self._str = intern(f"{self.target} := {self.expression}")


@define(eq=False)
class FieldAssign(Stmt):
    message: ID = field(converter=ID)
    field: ID = field(converter=ID)
    expression: Expr
    type_: rty.Message
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> FieldAssign:
        return FieldAssign(
            mapping.get(self.message, self.message),
            self.field,
            self.expression.substituted(mapping),
            self.type_,
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        target: z3.ExprRef
        if isinstance(self.expression, IntExpr):
            target = z3.Int(f"{self.message}.{self.field}")
        elif isinstance(self.expression, BoolExpr):
            target = z3.Bool(f"{self.message}.{self.field}")
        else:
            return z3.BoolVal(val=True)
        return target == self.expression.to_z3_expr()

    def _update_str(self) -> None:
        self._str = intern(f"{self.message}.{self.field} := {self.expression}")


@define(eq=False)
class Append(Stmt):
    sequence: ID = field(converter=ID)
    expression: Expr
    type_: rty.Sequence
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> Append:
        return Append(
            mapping.get(self.sequence, self.sequence),
            self.expression.substituted(mapping),
            self.type_,
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.BoolVal(val=True)

    def _update_str(self) -> None:
        self._str = intern(f"{self.sequence}'Append ({self.expression})")


@define(eq=False)
class Extend(Stmt):
    sequence: ID = field(converter=ID)
    expression: Expr
    type_: rty.Sequence
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> Extend:
        return Extend(
            mapping.get(self.sequence, self.sequence),
            self.expression.substituted(mapping),
            self.type_,
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.BoolVal(val=True)

    def _update_str(self) -> None:
        self._str = intern(f"{self.sequence}'Extend ({self.expression})")


@define(eq=False)
class Reset(Stmt):
    identifier: ID = field(converter=ID)
    parameter_values: Mapping[ID, Expr]
    type_: rty.Any
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> Reset:
        return Reset(
            mapping.get(self.identifier, self.identifier),
            {p: v.substituted(mapping) for p, v in self.parameter_values.items()},
            self.type_,
            self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.BoolVal(val=True)

    def _update_str(self) -> None:
        parameter_values = " => ".join(f"{i} => {e}" for i, e in self.parameter_values.items())
        if parameter_values:
            parameter_values = f" ({parameter_values})"
        self._str = intern(f"{self.identifier}'Reset{parameter_values}")


@define(eq=False)
class ChannelStmt(Stmt):
    channel: ID = field(converter=ID)
    expression: BasicExpr
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> ChannelStmt:
        return self.__class__(
            self.channel,
            self.expression.substituted(mapping),
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.BoolVal(val=True)

    def _update_str(self) -> None:
        self._str = intern(f"{self.channel}'{self.__class__.__name__} ({self.expression})")


@define(eq=False)
class Read(ChannelStmt):
    pass


@define(eq=False)
class Write(ChannelStmt):
    pass


@define(eq=False)
class Check(Stmt):
    expression: BoolExpr
    origin: Optional[Origin] = None

    @property
    def location(self) -> Optional[Location]:
        return self.expression.location

    def substituted(self, mapping: Mapping[ID, ID]) -> Check:
        return Check(
            self.expression.substituted(mapping),
            self.origin,
        )

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self.expression.preconditions(variable_id)

    def to_z3_expr(self) -> z3.BoolRef:
        return self.expression.to_z3_expr()

    def _update_str(self) -> None:
        self._str = intern(f"Check {self.expression}")


Self = TypeVar("Self", bound="Expr")


@define(init=False)
class Expr(Base):
    """Expression in three-address code (TAC) format."""

    origin: Optional[Origin]
    _str: Optional[str] = field(init=False, default=None)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return str(self) == str(other)
        return NotImplemented

    def __str__(self) -> str:
        if self._str is not None:
            return self._str

        self._update_str()
        assert self._str is not None
        return self._str  # type: ignore[unreachable]

    @property
    @abstractmethod
    def type_(self) -> rty.Any:
        raise NotImplementedError

    @property
    def origin_str(self) -> str:
        if self.origin:
            return str(self.origin)
        return str(self)

    @property
    def location(self) -> Optional[Location]:
        return self.origin.location if self.origin else None

    def substituted(self: Self, mapping: Mapping[ID, ID]) -> Self:
        raise NotImplementedError

    @abstractmethod
    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    @abstractmethod
    def _update_str(self) -> None:
        raise NotImplementedError


class BasicExpr(Expr):
    pass


class IntExpr(Expr):
    @property
    @abstractmethod
    def type_(self) -> rty.AnyInteger:
        raise NotImplementedError

    @abstractmethod
    def to_z3_expr(self) -> z3.ArithRef:
        raise NotImplementedError


class BoolExpr(Expr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    @abstractmethod
    def to_z3_expr(self) -> z3.BoolRef:
        raise NotImplementedError


class BasicIntExpr(BasicExpr, IntExpr):
    @property
    @abstractmethod
    def type_(self) -> rty.AnyInteger:
        raise NotImplementedError


class BasicBoolExpr(BasicExpr, BoolExpr):
    pass


@define(eq=False)
class Var(BasicExpr):
    identifier: ID = field(converter=ID)
    origin: Optional[Origin] = None

    def _update_str(self) -> None:
        self._str = intern(str(self.identifier))


@define(eq=False)
class IntVar(Var, BasicIntExpr):
    identifier: ID = field(converter=ID)
    var_type: rty.AnyInteger
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.AnyInteger:
        return self.var_type

    def substituted(self, mapping: Mapping[ID, ID]) -> IntVar:
        if self.identifier in mapping:
            return IntVar(mapping[self.identifier], self.var_type, self.origin)
        return self

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.Int(str(self.identifier))


@define(eq=False)
class BoolVar(Var, BasicBoolExpr):
    identifier: ID = field(converter=ID)
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def substituted(self, mapping: Mapping[ID, ID]) -> BoolVar:
        if self.identifier in mapping:
            return BoolVar(mapping[self.identifier], self.origin)
        return self

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Bool(str(self.identifier))


@define(eq=False)
class ObjVar(Var):
    identifier: ID = field(converter=ID)
    var_type: rty.Any
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        return self.var_type

    def substituted(self, mapping: Mapping[ID, ID]) -> ObjVar:
        if self.identifier in mapping:
            return ObjVar(mapping[self.identifier], self.var_type, self.origin)
        return self

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError


@define(eq=False)
class EnumLit(BasicExpr):
    identifier: ID = field(converter=ID)
    enum_type: rty.Enumeration
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Enumeration:
        return self.enum_type

    def substituted(self, _mapping: Mapping[ID, ID]) -> EnumLit:
        return self

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.Int(str(self.identifier))

    def _update_str(self) -> None:
        self._str = intern(str(self.identifier))


@define(eq=False)
class IntVal(BasicIntExpr):
    value: int
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.UniversalInteger:
        return rty.UniversalInteger(rty.Bounds(self.value, self.value))

    def substituted(self, _mapping: Mapping[ID, ID]) -> IntVal:
        return self

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.IntVal(self.value)

    def _update_str(self) -> None:
        self._str = intern(str(self.value))


@define(eq=False)
class BoolVal(BasicBoolExpr):
    value: bool
    origin: Optional[Origin] = None

    def substituted(self, _mapping: Mapping[ID, ID]) -> BoolVal:
        return self

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.BoolVal(val=self.value)

    def _update_str(self) -> None:
        self._str = intern(str(self.value))


@define(eq=False)
class Attr(Expr):
    prefix: ID = field(converter=ID)
    prefix_type: rty.Any
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> Attr:
        return self.__class__(
            mapping.get(self.prefix, self.prefix),
            self.prefix_type,
            self.origin,
        )

    def _update_str(self) -> None:
        symbol = re.sub(r"([a-z])([A-Z])", r"\1_\2", self.__class__.__name__)
        self._str = intern(f"{self.prefix}'{symbol}")


@define(eq=False)
class IntAttr(Attr, IntExpr):
    prefix: ID = field(converter=ID)
    prefix_type: rty.Any
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> IntAttr:
        return self.__class__(
            mapping.get(self.prefix, self.prefix),
            self.prefix_type,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.Int(str(self))


@define(eq=False)
class Size(IntAttr):
    @property
    def type_(self) -> rty.AnyInteger:
        return (
            rty.BIT_LENGTH
            if isinstance(self.prefix_type, (rty.Composite, rty.Compound))
            else rty.UniversalInteger()
        )


@define(eq=False)
class Length(IntAttr):
    @property
    def type_(self) -> rty.UniversalInteger:
        return rty.UniversalInteger()


@define(eq=False)
class First(IntAttr):
    @property
    def type_(self) -> rty.UniversalInteger:
        return rty.UniversalInteger()


@define(eq=False)
class Last(IntAttr):
    @property
    def type_(self) -> rty.UniversalInteger:
        return rty.UniversalInteger()


@define(eq=False)
class ValidChecksum(Attr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.ExprRef:
        return z3.Bool(str(self))


@define(eq=False)
class Valid(Attr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.ExprRef:
        return z3.Bool(str(self))


@define(eq=False)
class Present(Attr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.ExprRef:
        return z3.Bool(str(self))


@define(eq=False)
class HasData(Attr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.ExprRef:
        return z3.Bool(str(self))


@define(eq=False)
class Head(Attr):
    prefix: ID = field(converter=ID)
    prefix_type: rty.Composite
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        assert isinstance(self.prefix_type.element, rty.Any)
        return self.prefix_type.element

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError


@define(eq=False)
class Opaque(Attr):
    prefix: ID = field(converter=ID)
    prefix_type: Union[rty.Message, rty.Sequence]
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Sequence:
        return rty.OPAQUE

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError


@define(eq=False)
class FieldAccessAttr(Expr):
    message: ID = field(converter=ID)
    field: ID = field(converter=ID)
    message_type: rty.Compound
    origin: Optional[Origin] = None

    @property
    def field_type(self) -> rty.Any:
        type_ = self.message_type.field_types[self.field]
        assert isinstance(type_, rty.Any)
        return type_

    def substituted(self, mapping: Mapping[ID, ID]) -> FieldAccessAttr:
        return self.__class__(
            mapping.get(self.message, self.message),
            self.field,
            self.message_type,
            self.origin,
        )

    def _update_str(self) -> None:
        symbol = re.sub(r"([a-z])([A-Z])", r"\1_\2", self.__class__.__name__)
        self._str = intern(f"{self.message}.{self.field}'{symbol}")


@define(eq=False)
class FieldValid(FieldAccessAttr, BoolExpr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Bool(str(self))


@define(eq=False)
class FieldPresent(FieldAccessAttr, BoolExpr):
    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Bool(str(self))


@define(eq=False)
class FieldSize(FieldAccessAttr, IntExpr):
    @property
    def type_(self) -> rty.Integer:
        return rty.BIT_LENGTH

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.Int(str(self))


@define(eq=False)
class UnaryExpr(Expr):
    expression: BasicExpr
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> UnaryExpr:
        return self.__class__(self.expression.substituted(mapping), self.origin)


@define(eq=False)
class UnaryIntExpr(UnaryExpr, IntExpr):
    expression: BasicIntExpr
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.AnyInteger:
        return self.expression.type_


@define(eq=False)
class UnaryBoolExpr(UnaryExpr, BoolExpr):
    expression: BasicBoolExpr
    origin: Optional[Origin] = None


@define(eq=False)
class BinaryExpr(Expr):
    left: BasicExpr
    right: BasicExpr
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> BinaryExpr:
        return self.__class__(
            self.left.substituted(mapping),
            self.right.substituted(mapping),
            self.origin,
        )

    @property
    def origin_str(self) -> str:
        if self.origin is not None:
            return str(self.origin)
        return f"{self.left.origin_str}{self._symbol}{self.right.origin_str}"

    @property
    def location(self) -> Optional[Location]:
        if self.origin is not None:
            return self.origin.location
        if self.left.origin is not None:
            return self.left.origin.location
        return None

    def _update_str(self) -> None:
        self._str = intern(f"{self.left}{self._symbol}{self.right}")

    @property
    def _symbol(self) -> str:
        raise NotImplementedError


@define(eq=False)
class BinaryIntExpr(BinaryExpr, IntExpr):
    left: BasicIntExpr
    right: BasicIntExpr
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.AnyInteger:
        type_ = self.left.type_.common_type(self.right.type_)
        assert isinstance(type_, rty.AnyInteger)
        return type_


@define(eq=False)
class BinaryBoolExpr(BinaryExpr, BoolExpr):
    left: BasicBoolExpr
    right: BasicBoolExpr
    origin: Optional[Origin] = None


@define(eq=False)
class Neg(UnaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return -self.expression.to_z3_expr()

    def _update_str(self) -> None:
        self._str = intern(f"-{self.expression}")


@define(eq=False)
class Add(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() + self.right.to_z3_expr()

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        v_id = next(variable_id)
        v_type = to_integer(self.type_)
        upper_bound = (
            self.type_.bounds.upper
            if isinstance(self.type_, rty.BoundedInteger) and self.type_.bounds.upper is not None
            else INT_MAX
        )
        return [
            # Left + Right <= Upper_Bound
            Cond(
                LessEqual(
                    self.left,
                    IntVar(
                        v_id,
                        v_type,
                        origin=(
                            ConstructedOrigin(
                                f"{upper_bound} - {self.right.origin}",
                                self.origin.location,
                            )
                            if self.origin and self.right.origin
                            else None
                        ),
                    ),
                ),
                [
                    VarDecl(v_id, v_type, None, origin=self.origin),
                    Assign(
                        v_id,
                        Sub(
                            IntVal(upper_bound),
                            self.right,
                        ),
                        v_type,
                        origin=self.origin,
                    ),
                ],
            ),
        ]

    @property
    def _symbol(self) -> str:
        return " + "


@define(eq=False)
class Sub(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() - self.right.to_z3_expr()

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return [
            # Left >= Right
            Cond(GreaterEqual(self.left, self.right)),
        ]

    @property
    def _symbol(self) -> str:
        return " - "


@define(eq=False)
class Mul(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() * self.right.to_z3_expr()

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        v_id = next(variable_id)
        v_type = to_integer(self.type_)
        upper_bound = (
            self.type_.bounds.upper
            if isinstance(self.type_, rty.BoundedInteger) and self.type_.bounds.upper is not None
            else INT_MAX
        )
        return [
            # Left * Right <= Upper_Bound
            Cond(
                LessEqual(
                    self.left,
                    IntVar(
                        v_id,
                        v_type,
                        origin=(
                            ConstructedOrigin(
                                f"{upper_bound} / {self.right.origin}",
                                self.origin.location,
                            )
                            if self.origin and self.right.origin
                            else None
                        ),
                    ),
                ),
                [
                    VarDecl(v_id, v_type, None, origin=self.origin),
                    Assign(v_id, Div(IntVal(upper_bound), self.right), v_type, origin=self.origin),
                ],
            ),
        ]

    @property
    def _symbol(self) -> str:
        return " * "


@define(eq=False)
class Div(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() / self.right.to_z3_expr()

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return [
            # Right /= 0
            Cond(NotEqual(self.right, IntVal(0))),
        ]

    @property
    def _symbol(self) -> str:
        return " / "


@define(eq=False)
class Pow(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() ** self.right.to_z3_expr()

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        v_id = next(variable_id)
        v_type = to_integer(self.type_)
        upper_bound = (
            self.type_.bounds.upper
            if isinstance(self.type_, rty.BoundedInteger) and self.type_.bounds.upper is not None
            else INT_MAX
        )
        return [
            # Left ** Right <= Upper_Bound
            Cond(
                LessEqual(
                    IntVar(v_id, self.type_, origin=self.origin),
                    IntVal(upper_bound),
                ),
                [
                    VarDecl(v_id, v_type, None, origin=self.origin),
                    Assign(v_id, self, to_integer(self.type_), origin=self.origin),
                ],
            ),
        ]

    @property
    def _symbol(self) -> str:
        return " ** "


@define(eq=False)
class Mod(BinaryIntExpr):
    def to_z3_expr(self) -> z3.ArithRef:
        return self.left.to_z3_expr() % self.right.to_z3_expr()

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return [
            # Right /= 0
            Cond(NotEqual(self.right, IntVal(0))),
        ]

    @property
    def _symbol(self) -> str:
        return " mod "


@define(eq=False)
class Not(UnaryBoolExpr):
    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Not(self.expression.to_z3_expr())

    def _update_str(self) -> None:
        self._str = intern(f"not {self.expression}")


@define(eq=False)
class And(BinaryBoolExpr):
    def to_z3_expr(self) -> z3.BoolRef:
        return z3.And(self.left.to_z3_expr(), self.right.to_z3_expr())

    @property
    def _symbol(self) -> str:
        return " and "


@define(eq=False)
class Or(BinaryBoolExpr):
    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Or(self.left.to_z3_expr(), self.right.to_z3_expr())

    @property
    def _symbol(self) -> str:
        return " or "


@define(eq=False)
class Relation(BoolExpr, BinaryExpr):
    left: BasicExpr
    right: BasicExpr
    origin: Optional[Origin] = None


@define(eq=False)
class Less(Relation):
    left: BasicIntExpr
    right: BasicIntExpr
    origin: Optional[Origin] = None

    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() < self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " < "


@define(eq=False)
class LessEqual(Relation):
    left: BasicIntExpr
    right: BasicIntExpr
    origin: Optional[Origin] = None

    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() <= self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " <= "


@define(eq=False)
class Equal(Relation):
    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() == self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " = "


@define(eq=False)
class GreaterEqual(Relation):
    left: BasicIntExpr
    right: BasicIntExpr
    origin: Optional[Origin] = None

    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() >= self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " >= "


@define(eq=False)
class Greater(Relation):
    left: BasicIntExpr
    right: BasicIntExpr
    origin: Optional[Origin] = None

    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() > self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " > "


@define(eq=False)
class NotEqual(Relation):
    def to_z3_expr(self) -> z3.BoolRef:
        return self.left.to_z3_expr() != self.right.to_z3_expr()

    @property
    def _symbol(self) -> str:
        return " /= "


@define(eq=False)
class Call(Expr):
    identifier: ID = field(converter=ID)
    arguments: Sequence[Expr]
    argument_types: Sequence[rty.Any]
    origin: Optional[Origin] = None
    _preconditions: list[Cond] = field(init=False, factory=list)

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return self._preconditions

    def set_preconditions(self, preconditions: list[Cond]) -> None:
        self._preconditions = preconditions

    def _update_str(self) -> None:
        self._str = intern(
            str(self.identifier)
            + (f" ({', '.join(str(a) for a in self.arguments)})" if self.arguments else ""),
        )


@define(eq=False)
class IntCall(Call, IntExpr):
    identifier: ID = field(converter=ID)
    arguments: Sequence[Expr]
    argument_types: Sequence[rty.Any]
    type_: rty.AnyInteger
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> IntCall:
        return self.__class__(
            self.identifier,
            [a.substituted(mapping) for a in self.arguments],
            self.argument_types,
            self.type_,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ArithRef:
        # TODO(eng/recordflux/RecordFlux#1338): Return value need not to be identical
        # The return value may be different even if the arguments are the same.
        return z3.Int(str(self.identifier))

    def _update_str(self) -> None:
        self._str = intern(
            str(self.identifier)
            + (f" ({', '.join(str(a) for a in self.arguments)})" if self.arguments else ""),
        )


@define(eq=False)
class BoolCall(Call, BoolExpr):
    def substituted(self, mapping: Mapping[ID, ID]) -> BoolCall:
        return self.__class__(
            self.identifier,
            [a.substituted(mapping) for a in self.arguments],
            self.argument_types,
            self.origin,
        )

    def to_z3_expr(self) -> z3.BoolRef:
        # TODO(eng/recordflux/RecordFlux#1338): Return value need not to be identical
        # The return value may be different even if the arguments are the same.
        return z3.Bool(str(self.identifier))


@define(eq=False)
class ObjCall(Call):
    identifier: ID = field(converter=ID)
    arguments: Sequence[Expr]
    argument_types: Sequence[rty.Any]
    type_: rty.Any
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> ObjCall:
        return self.__class__(
            self.identifier,
            [a.substituted(mapping) for a in self.arguments],
            self.argument_types,
            self.type_,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError


@define(eq=False)
class FieldAccess(Expr):
    message: ID = field(converter=ID)
    field: ID = field(converter=ID)
    message_type: rty.Compound
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> FieldAccess:
        return self.__class__(
            mapping.get(self.message, self.message),
            self.field,
            self.message_type,
            self.origin,
        )

    def _update_str(self) -> None:
        self._str = intern(f"{self.message}.{self.field}")


@define(eq=False)
class IntFieldAccess(FieldAccess, IntExpr):
    message: ID = field(converter=ID)
    field: ID = field(converter=ID)
    message_type: rty.Compound
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.AnyInteger:
        type_ = self.message_type.types[self.field]
        assert isinstance(type_, rty.AnyInteger)
        return type_

    def substituted(self, mapping: Mapping[ID, ID]) -> IntFieldAccess:
        return self.__class__(
            mapping.get(self.message, self.message),
            self.field,
            self.message_type,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ArithRef:
        return z3.Int(str(self))

    def _update_str(self) -> None:
        self._str = intern(f"{self.message}.{self.field}")


@define(eq=False)
class BoolFieldAccess(FieldAccess, BoolExpr):
    def to_z3_expr(self) -> z3.BoolRef:
        return z3.Bool(str(self))


@define(eq=False)
class ObjFieldAccess(FieldAccess):
    message: ID = field(converter=ID)
    field: ID = field(converter=ID)
    message_type: rty.Compound
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        type_ = self.message_type.field_types[self.field]
        assert isinstance(type_, rty.Any)
        return type_

    def substituted(self, mapping: Mapping[ID, ID]) -> ObjFieldAccess:
        return self.__class__(
            mapping.get(self.message, self.message),
            self.field,
            self.message_type,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError


@define(eq=False)
class IfExpr(Expr):
    condition: BasicBoolExpr
    then_expr: ComplexExpr
    else_expr: ComplexExpr
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> IfExpr:
        return self.__class__(
            self.condition.substituted(mapping),
            self.then_expr.substituted(mapping),
            self.else_expr.substituted(mapping),
            self.origin,
        )

    def to_z3_expr(self) -> z3.ExprRef:
        # TODO(eng/recordflux/RecordFlux#1339): Fix handling of complex expressions
        return z3.If(
            self.condition.to_z3_expr(),
            self.then_expr.expr.to_z3_expr(),
            self.else_expr.expr.to_z3_expr(),
        )

    def _update_str(self) -> None:
        self._str = intern(f"(if {self.condition} then {self.then_expr} else {self.else_expr})")


@define(eq=False)
class IntIfExpr(IfExpr, IntExpr):
    condition: BasicBoolExpr
    then_expr: ComplexIntExpr
    else_expr: ComplexIntExpr
    return_type: rty.AnyInteger
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.AnyInteger:
        return self.return_type

    def to_z3_expr(self) -> z3.ArithRef:
        result = super().to_z3_expr()
        assert isinstance(result, z3.ArithRef)
        return result


@define(eq=False)
class BoolIfExpr(IfExpr, BoolExpr):
    condition: BasicBoolExpr
    then_expr: ComplexBoolExpr
    else_expr: ComplexBoolExpr
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Enumeration:
        return rty.BOOLEAN

    def to_z3_expr(self) -> z3.BoolRef:
        result = super().to_z3_expr()
        assert isinstance(result, z3.BoolRef)
        return result


@define(eq=False)
class Conversion(Expr):
    identifier: ID = field(converter=ID)
    argument: Expr
    target_type: rty.Any
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        return self.target_type

    def substituted(self, mapping: Mapping[ID, ID]) -> Conversion:
        return self.__class__(
            self.identifier,
            self.argument.substituted(mapping),
            self.target_type,
            self.origin,
        )

    def to_z3_expr(self) -> z3.ExprRef:
        return z3.BoolVal(val=True)

    def _update_str(self) -> None:
        self._str = intern(f"{self.identifier} ({self.argument})")


@define(eq=False)
class IntConversion(Conversion, BasicIntExpr):
    identifier: ID = field(converter=ID)
    argument: IntExpr
    target_type: rty.Integer
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Integer:
        return self.target_type

    def to_z3_expr(self) -> z3.ArithRef:
        return self.argument.to_z3_expr()


@define(eq=False)
class Comprehension(Expr):
    iterator: ID = field(converter=ID)
    sequence: Union[Var, FieldAccess]
    selector: ComplexExpr
    condition: ComplexBoolExpr
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Aggregate:
        return rty.Aggregate(self.selector.expr.type_)

    def substituted(self, mapping: Mapping[ID, ID]) -> Comprehension:
        return self.__class__(
            mapping.get(self.iterator, self.iterator),
            self.sequence.substituted(mapping),
            self.selector.__class__(
                [s.substituted(mapping) for s in self.selector.stmts],
                self.selector.expr.substituted(mapping),
            ),
            self.condition.__class__(
                [s.substituted(mapping) for s in self.condition.stmts],
                self.condition.expr.substituted(mapping),
            ),
            self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        self._str = intern(
            f"[for {self.iterator} in {self.sequence} if {self.condition} => {self.selector}]",
        )


@define(eq=False)
class Find(Expr):
    iterator: ID = field(converter=ID)
    sequence: Union[Var, FieldAccess]
    selector: ComplexExpr
    condition: ComplexBoolExpr
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        return self.selector.expr.type_

    def substituted(self, mapping: Mapping[ID, ID]) -> Find:
        return self.__class__(
            mapping.get(self.iterator, self.iterator),
            self.sequence.substituted(mapping),
            self.selector.__class__(
                [s.substituted(mapping) for s in self.selector.stmts],
                self.selector.expr.substituted(mapping),
            ),
            self.condition.__class__(
                [s.substituted(mapping) for s in self.condition.stmts],
                self.condition.expr.substituted(mapping),
            ),
            self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        self._str = intern(
            f"Find (for {self.iterator} in {self.sequence} if {self.condition} => {self.selector})",
        )


@define(eq=False)
class Agg(Expr):
    elements: Sequence[BasicExpr]
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Aggregate:
        return rty.Aggregate(rty.common_type([e.type_ for e in self.elements]))

    def substituted(self, mapping: Mapping[ID, ID]) -> Agg:
        return self.__class__(
            [e.substituted(mapping) for e in self.elements],
            origin=self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        self._str = intern("[" + ", ".join(map(str, self.elements)) + "]")


def _named_agg_elements_converter(
    elements: Sequence[tuple[Union[StrID, BasicExpr], BasicExpr]],
) -> Sequence[tuple[Union[ID, BasicExpr], BasicExpr]]:
    return [(ID(n) if isinstance(n, str) else n, e) for n, e in elements]


@define(eq=False)
class NamedAgg(Expr):
    """Only used by code generator and therefore provides minimum functionality."""

    elements: Sequence[tuple[Union[ID, BasicExpr], BasicExpr]] = field(
        converter=_named_agg_elements_converter,
    )
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Any:
        raise NotImplementedError

    def substituted(self, mapping: Mapping[ID, ID]) -> NamedAgg:
        raise NotImplementedError

    def preconditions(self, variable_id: Generator[ID, None, None]) -> list[Cond]:
        raise NotImplementedError

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        self._str = intern("[" + ", ".join([f"{k} => {v}" for k, v in self.elements]) + "]")


@define(eq=False)
class Str(Expr):
    string: str
    origin: Optional[Origin] = None

    @property
    def type_(self) -> rty.Sequence:
        return rty.OPAQUE

    def substituted(self, _mapping: Mapping[ID, ID]) -> Str:
        return self

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        self._str = intern(f'"{self.string}"')


@define(eq=False)
class MsgAgg(Expr):
    identifier: ID = field(converter=ID)
    field_values: Mapping[ID, Expr]
    type_: rty.Message
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> MsgAgg:
        return self.__class__(
            self.identifier,
            {f: v.substituted(mapping) for f, v in self.field_values.items()},
            self.type_,
            origin=self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        field_values = (
            " => ".join(f"{i} => {e}" for i, e in self.field_values.items())
            if self.field_values
            else "null message"
        )
        self._str = intern(f"{self.identifier}'({field_values})")


@define(eq=False)
class DeltaMsgAgg(Expr):
    identifier: ID = field(converter=ID)
    field_values: Mapping[ID, Expr]
    type_: rty.Message
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> DeltaMsgAgg:
        return self.__class__(
            self.identifier,
            {f: v.substituted(mapping) for f, v in self.field_values.items()},
            self.type_,
            origin=self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        field_values = (
            " => ".join(f"{i} => {e}" for i, e in self.field_values.items())
            if self.field_values
            else "null message"
        )
        self._str = intern(f"{self.identifier} with delta {field_values}")


@define(eq=False)
class CaseExpr(Expr):
    expression: BasicExpr
    choices: Sequence[tuple[Sequence[BasicExpr], BasicExpr]]
    type_: rty.Any
    origin: Optional[Origin] = None

    def substituted(self, mapping: Mapping[ID, ID]) -> CaseExpr:
        return self.__class__(
            self.expression.substituted(mapping),
            [
                (
                    [v.substituted(mapping) for v in vs],
                    e.substituted(mapping),
                )
                for vs, e in self.choices
            ],
            self.type_,
            self.origin,
        )

    def preconditions(self, _variable_id: Generator[ID, None, None]) -> list[Cond]:
        return []

    def to_z3_expr(self) -> z3.ExprRef:
        raise NotImplementedError

    def _update_str(self) -> None:
        data = ",\n".join(f"      when {' | '.join(map(str, c))} => {e}" for c, e in self.choices)
        self._str = intern(f"(case {self.expression} is\n{data})")


@frozen
class Decl:
    identifier: ID = field(converter=ID)
    location: Optional[Location]


@frozen
class FormalDecl(Decl):
    pass


@frozen
class Argument:
    identifier: ID = field(converter=ID)
    type_identifier: ID = field(converter=ID)
    type_: rty.Type


@frozen
class FuncDecl(FormalDecl):
    identifier: ID = field(converter=ID)
    arguments: Sequence[Argument]
    return_type: ID = field(converter=ID)
    type_: rty.Type
    location: Optional[Location]


@frozen
class ChannelDecl(FormalDecl):
    identifier: ID = field(converter=ID)
    readable: bool
    writable: bool
    location: Optional[Location]


@frozen
class ComplexExpr(Base):
    stmts: list[Stmt]
    expr: Expr

    def __str__(self) -> str:
        statements = "".join([f"{s}; " for s in self.stmts])
        return f"{{{statements}{self.expr}}}"

    def is_expr(self) -> bool:
        return len(self.stmts) == 0

    def is_basic_expr(self) -> bool:
        return self.is_expr() and isinstance(self.expr, BasicExpr)

    def substituted(self, mapping: Mapping[ID, ID]) -> ComplexExpr:
        return self.__class__(
            [s.substituted(mapping) for s in self.stmts],
            self.expr.substituted(mapping),
        )


@frozen
class ComplexIntExpr(ComplexExpr):
    stmts: list[Stmt]
    expr: IntExpr


@frozen
class ComplexBoolExpr(ComplexExpr):
    stmts: list[Stmt]
    expr: BoolExpr


@frozen
class Transition:
    target: ID = field(converter=ID)
    condition: ComplexExpr
    description: Optional[str]
    location: Optional[Location]


@frozen
class State:
    identifier: ID = field(converter=ID)
    transitions: Sequence[Transition]
    exception_transition: Optional[Transition]
    actions: Sequence[Stmt]
    description: Optional[str]
    location: Optional[Location]

    @property
    def declarations(self) -> list[VarDecl]:
        return [a for a in self.actions if isinstance(a, VarDecl)]


FINAL_STATE = State("Final", [], None, [], None, None)


@frozen(init=False)
class Session:
    identifier: ID = field(converter=ID)
    states: Sequence[State]
    declarations: Sequence[VarDecl]
    parameters: Sequence[FormalDecl]
    types: Mapping[ID, mty.Type]
    location: Optional[Location]

    def __init__(  # noqa: PLR0913
        self,
        identifier: ID,
        states: Sequence[State],
        declarations: Sequence[VarDecl],
        parameters: Sequence[FormalDecl],
        types: Mapping[ID, mty.Type],
        location: Optional[Location],
        variable_id: Generator[ID, None, None],
        workers: int = 1,
    ) -> None:
        manager = ProofManager(workers)
        states = [
            State(
                s.identifier,
                [
                    Transition(
                        t.target,
                        ComplexExpr(
                            add_required_checks(t.condition.stmts, manager, variable_id),
                            t.condition.expr,
                        ),
                        t.description,
                        t.location,
                    )
                    for t in s.transitions
                ],
                s.exception_transition,
                add_required_checks(s.actions, manager, variable_id),
                s.description,
                s.location,
            )
            for s in states
        ]
        self.__attrs_init__(identifier, states, declarations, parameters, types, location)

    @property
    def package(self) -> ID:
        return self.identifier.parent

    @property
    def initial_state(self) -> State:
        return self.states[0]

    @identifier.validator
    def _check_identifier(self, attribute: str, value: ID) -> None:
        assert len(value.parts) == 2, attribute


def add_checks(statements: Sequence[Stmt], variable_id: Generator[ID, None, None]) -> list[Stmt]:
    """Add check statements in places where preconditions must be ensured."""

    result = []

    for statement in statements:
        preconditions = statement.preconditions(variable_id)

        for precondition in preconditions:
            result.extend([*precondition.facts, Check(precondition.goal)])

        result.append(statement)

    return result


def remove_unnecessary_checks(statements: Sequence[Stmt], manager: ProofManager) -> list[Stmt]:
    """Remove all checks that are always true."""

    always_true: list[int] = []

    facts: list[Stmt] = []

    for i, s in enumerate(statements):
        if isinstance(s, Check):
            manager.add(
                [
                    ProofJob(
                        [
                            Check(Not(BoolVar("__GOAL__"))),
                            Assign("__GOAL__", s.expression, s.expression.type_),
                            *facts,
                        ],
                        {
                            ProofResult.UNSAT: [i],
                            ProofResult.SAT: [],
                            ProofResult.UNKNOWN: [],
                        },
                    ),
                ],
            )
            facts.append(s)

    results = manager.check()

    for r in results:
        always_true.extend(r.result)

    result = list(statements)
    for i in reversed(always_true):
        result = [*result[:i], *result[i + 1 :]]

    return remove_unused_assignments(result)


def remove_unused_assignments(statements: Sequence[Stmt]) -> list[Stmt]:
    # TODO(eng/recordflux/RecordFlux#1339): Add removal of unused assignments
    return list(statements)


def add_required_checks(
    statements: Sequence[Stmt],
    manager: ProofManager,
    variable_id: Generator[ID, None, None],
) -> list[Stmt]:
    """
    Add check statements in places where preconditions are not always true.

    For each statement it is checked, if its preconditions are always true. If this is not the
    case, a check statement is added in front of the respective statement. The check statements
    in the resulting list mark the places where the code generator must insert explicit checks.
    """
    result = remove_unnecessary_checks(add_checks(statements, variable_id), manager)

    for s in result:
        if isinstance(s, Check):
            info(
                f'precondition "{s.expression.origin_str}" must be checked at runtime',
                Subsystem.MODEL,
                s.expression.location,
            )

    return result


def to_integer(type_: rty.AnyInteger) -> rty.Integer:
    return type_ if isinstance(type_, rty.Integer) else rty.BASE_INTEGER



from __future__ import annotations

import textwrap
from dataclasses import dataclass, field

from rflx import const
from rflx.identifier import ID

from . import top_level_declaration


@dataclass
class Package:
    name: ID
    imports: frozenset[ID] = field(default_factory=frozenset)
    declarations: list[top_level_declaration.TopLevelDeclaration] = field(default_factory=list)

    @property
    def imports_str(self) -> str:
        explicit_imports = self.imports - {
            self.name,
            const.BUILTINS_PACKAGE,
            const.INTERNAL_PACKAGE,
        }
        return "\n".join(f"with {i};" for i in sorted(explicit_imports))

    @property
    def begin_str(self) -> str:
        return f"package {self.name} is"

    @property
    def end_str(self) -> str:
        return f"end {self.name};"

    @property
    def declarations_str(self) -> str:
        raw = "\n\n".join(f"{t};" for t in self.declarations)
        return textwrap.indent(raw, " " * 3)

    def __str__(self) -> str:
        pieces = [self.imports_str, self.begin_str, self.declarations_str, self.end_str]
        return "\n\n".join(filter(None, pieces))
